# -*- coding: utf-8 -*-

from loguru import logger

from mozci.console.commands.push import BasePushCommand


class RegressionCommand(BasePushCommand):
    name = "regression"
    description = (
        "Lists tasks with potential regression, that may introduce a build bustage"
    )

    def handle(self):
        for push in self.pushes:
            tasks_to_retrigger = push.check_build_regressions()
            if not tasks_to_retrigger:
                logger.info("No build task detected as potential regression")
                return
            logger.info(
                f"{len(tasks_to_retrigger)} tasks have been detected as potential regressions:"
            )
            for task in tasks_to_retrigger:
                logger.info(f" * {task.label} [{task.id}]")



# -*- coding: utf-8 -*-

import threading
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Dict, List

from lru import LRU

from mozci.data.base import DataSource
from mozci.errors import ContractNotFilled
from mozci.task import Task
from mozci.util.memoize import memoized_property
from mozci.util.req import get_session

try:
    from treeherder.model.models import Job, Push
except ImportError:
    Job = None
    Push = None


class JobUnavailable(Exception):
    """Exception raised when a Treeherder job does not exist or its logs have not been parsed yet"""


class BaseTreeherderSource(DataSource, ABC):
    lock = threading.Lock()
    groups_cache: Dict[str, List[str]] = LRU(7000)

    @abstractmethod
    def get_push_test_groups(self, branch: str, rev: str) -> Dict[str, List[str]]: ...

    def run_test_task_groups(self, branch, rev, task):
        # Use a lock since push.py invokes this across many threads (which is
        # useful for the 'errorsummary' data source, but not here). This ensures
        # we don't make more than one request to Treeherder.
        with self.lock:
            if task.id not in self.groups_cache:
                self.groups_cache.update(self.get_push_test_groups(branch, rev))

        try:
            # TODO: Once https://github.com/mozilla/mozci/issues/662 is fixed, we should return the actual duration instead of None.
            return {
                group: (status, None)
                for group, status in self.groups_cache.pop(task.id).items()
            }
        except KeyError:
            raise ContractNotFilled(self.name, "test_task_groups", "groups are missing")


class TreeherderClientSource(BaseTreeherderSource):
    """Uses the public API to query Treeherder."""

    name = "treeherder_client"
    supported_contracts = ("push_tasks_classifications", "test_task_groups", "pushes")
    base_url = "https://treeherder.mozilla.org/api"

    @memoized_property
    def session(self):
        session = get_session()
        session.headers = {"User-Agent": "mozci"}
        return session

    @lru_cache(maxsize=700)
    def _run_query(self, query, params=None):
        query = query.lstrip("/")
        url = f"{self.base_url}/{query}"

        params = params or {}
        params.setdefault("format", "json")

        response = self.session.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def run_push_tasks_classifications(self, branch, rev):
        data = self._run_query(f"/project/{branch}/note/push_notes/?revision={rev}")

        classifications = {}
        for item in data:
            job = item["job"]
            classifications[job["task_id"]] = {
                "classification": item["failure_classification_name"]
            }
            if item["text"]:
                classifications[job["task_id"]]["classification_note"] = item["text"]
        return classifications

    def get_push_test_groups(self, branch, rev) -> Dict[str, List[str]]:
        data = self._run_query(f"/project/{branch}/push/group_results/?revision={rev}")
        return {k: v for k, v in data.items() if v if k not in ("", "default")}

    def run_pushes(self, branch, nb=15) -> List[Dict]:
        response = self._run_query(f"/project/{branch}/push/?count={nb}")
        return [
            {
                "pushid": p["id"],
                "date": p["push_timestamp"],
                "revs": [r["revision"] for r in p["revisions"]],
            }
            for p in response["results"]
        ]

    def get_job_from_task(self, task: Task) -> dict:
        """Map a task to its Treeherder job, based on the Job list API.
        The endpoint returns a list of jobs, but only 1 result should be present using the `task_id` parameter.
        """
        data = self._run_query(f"/jobs/?task_id={task.id}")
        job = next(iter(data.get("results", [])), None)
        if not job:
            raise JobUnavailable
        if "job_property_names" not in data:
            raise Exception(
                "`job_property_names` is missing from /jobs/ treeherder API response"
            )
        return {key: value for key, value in zip(data["job_property_names"], job)}

    def check_job_ready(self, job_id: int, branch: str = "autoland") -> None:
        """Check if a specific job has been processed in Treeherder already.
        Raises JobUnavailable in case no job log are found or some are still in a pending state.
        """
        data = self._run_query(f"/project/{branch}/job-log-url/?job_id={job_id}")
        if (
            not data
            # Job logs can have final state "parsed", "failed" or "skipped_size"
            or any(job_log_url.get("parse_status") == "pending" for job_log_url in data)
        ):
            raise JobUnavailable

    def get_bug_suggestions(self, job_id: int, branch: str = "autoland") -> list[dict]:
        """List suggestions for a specific job."""
        return self._run_query(f"/project/{branch}/jobs/{job_id}/bug_suggestions/")



# -*- coding: utf-8 -*-
from __future__ import annotations

import collections
import fnmatch
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from inspect import signature
from statistics import median
from typing import Dict, List, NewType, Optional, Tuple

import jsone
import requests
import taskcluster
import yaml
from loguru import logger

from mozci import config, data
from mozci.errors import ArtifactNotFound, TaskNotFound
from mozci.util.defs import INTERMITTENT_CLASSES
from mozci.util.memoize import memoize, memoized_property
from mozci.util.taskcluster import (
    PRODUCTION_TASKCLUSTER_ROOT_URL,
    find_task_id,
    get_artifact,
    list_artifacts,
)


class Status(Enum):
    PASS = 0
    FAIL = 1
    INTERMITTENT = 2


@memoize
def fetch_test_variant_yaml() -> Dict:
    # load variants.yml from mozilla-central
    test_variants = {}
    try:
        r = requests.get(
            "https://hg.mozilla.org/mozilla-central/raw-file/default/taskcluster/test_configs/variants.yml"
        )
        tv = yaml.load(r.text, yaml.BaseLoader)
        for v in tv:
            test_variants[v] = tv[v]["suffix"]
    except requests.ConnectionError:
        raise
    return test_variants


NO_GROUPS_SUITES = (
    "raptor",
    "talos",
    "awsy",
    "gtest",
    "cppunit",
    "telemetry-tests",
    "firefox-ui-functional",
    "junit",  # https://bugzilla.mozilla.org/show_bug.cgi?id=1617632
    "jittest",  # https://bugzilla.mozilla.org/show_bug.cgi?id=1617633
    "marionette",  # https://bugzilla.mozilla.org/show_bug.cgi?id=1636088
)


def is_no_groups_suite(label):
    return any(f"-{s}-" in label for s in NO_GROUPS_SUITES)


slash_group_warned = False


def is_bad_group(task_id: str, group: str) -> bool:
    bad_group = (
        not group.strip()
        or group.startswith("file://")
        or group.startswith("Z:")
        or os.path.isabs(group)
        or "\\" in group
    )

    if bad_group:
        logger.error(f"Bad group name in task {task_id}: '{group}'")

    return bad_group


def is_autoclassifiable(task: TestTask) -> bool:
    """Check a task is enabled for auto-classification
    by applying glob patterns from configuration
    """
    assert task.label, "Missing task label"

    if not config["autoclassification"]["enabled"]:
        return False

    allowed_values = set(failure_type.value for failure_type in FailureType)
    filtered_failure_types = config["autoclassification"]["failure-types"]
    assert isinstance(filtered_failure_types, list) and set(
        filtered_failure_types
    ).issubset(allowed_values), "Unsupported failure types in configuration"

    flat_failure_types = list(
        set(
            test_and_type
            for group in task.failure_types.values()
            for test_and_type in group
        )
    )

    return (
        any(
            fnmatch.fnmatch(task.label, pattern)
            for pattern in config["autoclassification"]["test-suite-names"]
        )
        and len(flat_failure_types) == 1
        and flat_failure_types[0][1].value in filtered_failure_types
    )


# Transform WPT group names to a full relative path in mozilla-central.
def wpt_workaround(group: str) -> str:
    # No need to transform empty groups (also, they will be filtered out
    # in a following step).
    if not group.strip():
        return group

    if group.startswith(":"):
        assert group.startswith(":/"), (
            f"Group {group} starts with : but doesn't start with :/"
        )
        group = group[1:]
    assert group.startswith("/"), f"Group {group} doesn't start with /"
    if group.startswith("/_mozilla/"):
        return "/".join(
            ["testing/web-platform/mozilla/tests", group[len("/_mozilla/") :]]
        )
    else:
        return "/".join(["testing/web-platform/tests", group[1:]])


def get_test_variant(variants, label) -> str:
    if not variants:
        return ""

    suffixes = []
    yaml_variants = fetch_test_variant_yaml()
    for variant in variants:
        if variant not in yaml_variants:
            raise Exception(
                f"Missing test variant {variant} in variants.yml: {yaml_variants}"
            )
        suffixes.append(yaml_variants[variant])

    # figure out which order to pack the suffixes; we can cheat by using self.label
    retVal = sorted(suffixes, key=lambda x: label.find(x))
    return "-".join(retVal)


def get_configuration(task) -> str:
    platform = task["task"]["extra"]["treeherder-platform"]
    suite = task["task"]["extra"]["suite"]
    label = task["label"]
    test_variant = get_test_variant(
        task["task"]["extra"].get("test-settings", {}).get("runtime", {}), label
    )
    return _get_configuration(platform, suite, label, test_variant)


def _get_configuration(platform, suite, label, test_variant) -> str:
    assert platform is not None
    retVal = f"test-{platform}"

    # Hack #1: Android platform names do not match
    if "android" in platform:
        platform = platform.replace("7-0", "7.0")
        retVal = f"test-{platform}"
        if suite is None:
            suite = "junit"
        if "junit" not in suite:
            retVal += "-geckoview"

    # account for the suite being replaced
    retVal += "-*"

    # Hack #2: wpt tasks using tag=<X> have no attributes in test-settings
    wpt_tags = ["privatebrowsing", "canvas", "webgpu"]
    for wt in wpt_tags:
        if f"web-platform-tests-{wt}" in label:
            retVal += f"-{wt}"

    # Hack #3: xpcshell-msix is a separate suite, but still has test_variant listed
    # remove this on April 15, 2025 as that is 1 year after bug 1891629 landed
    if test_variant:
        if not (suite == "xpcshell-msix" and test_variant == "msix"):
            retVal += f"-{test_variant}"

    return retVal


@dataclass
class Task:
    """Contains information pertaining to a single task."""

    id: str
    label: Optional[str] = field(default=None)
    duration: Optional[int] = field(default=None)
    result: Optional[str] = field(default=None)
    state: Optional[str] = field(default=None)
    queue_id: Optional[str] = field(default=None)
    action: Optional[Dict] = field(default=None)
    parent: Optional[str] = field(default=None)
    classification: Optional[str] = field(default="not classified")
    classification_note: Optional[str] = field(default=None)
    tags: Dict = field(default_factory=dict)
    tier: Optional[int] = field(default=None)
    suite: Optional[str] = field(default=None)
    platform: Optional[str] = field(default=None)
    variant: Optional[Dict] = field(default=None)
    job_kind: Optional[str] = field(default=None)

    @staticmethod
    def create(index=None, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL, **kwargs):
        """Factory method to create a new Task instance.

        One of ``index`` or ``id`` must be specified.

        Args:
            index (str): Taskcluster index path used to find the task id (optional).
            kwargs (dict): Arguments to forward to the :class:`~mozci.task.Task` constructor.

        Raises:
            :class:`~mozci.errors.TaskNotFound`: when the task identified by
                specified index or task id could not be found.
        """
        if index and "id" not in kwargs:
            try:
                kwargs["id"] = find_task_id(index, root_url=root_url)
            except requests.exceptions.HTTPError as e:
                label = kwargs.get("label", "unknown label")
                raise TaskNotFound(id=index, label=label) from e

        if kwargs.get("label", "").startswith("test-"):
            return TestTask(**kwargs)
        return Task(**kwargs)

    @property
    def is_backfill(self) -> bool:
        return self.tags.get("action", "") == "backfill-task"

    @property
    def is_retrigger(self) -> bool:
        return self.tags.get("action", "").startswith("retrigger-")

    @property
    def is_tests_grouped(self) -> bool:
        return self.tags.get("tests_grouped", "") == "1"

    @property
    def failed(self):
        """Tasks with result `deadline-exceeded` are not considered as
        failures because it is likely that a dependency did not run.
        """
        return self.result in ("busted", "failed", "exception")

    @property
    def artifacts(self):
        """List the artifacts that were uploaded by this task."""
        return [artifact["name"] for artifact in list_artifacts(self.id)]

    @property
    def configuration(self):
        return _get_configuration(
            self.platform,
            self.suite,
            self.label,
            get_test_variant(self.variant, self.label),
        )

    def get_artifact(self, path, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL):
        """Downloads and returns the content of an artifact.

        Args:
            path (str): The path component of the artifact URL. This is usually
                        something like `public/file.txt`. The values listed by
                        the `Task.artifacts` property can be passed into this
                        function.

        Returns:
            Contents of the artifact.

        Raises:
            :class:`~mozci.errors.ArtifactNotFound`: When the requested
                artifact does not exist.
        """
        try:
            data = get_artifact(self.id, path, root_url=root_url)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                raise ArtifactNotFound(path, self.id, self.label) from e
            raise

        return data

    def to_json(self):
        """A JSON compatible representation of this Task in dictionary form.

        Only values passed in to the constructor will be included.

        Returns:
            dict: A JSON-compatible representation of the task.
        """
        sig = signature(self.__init__)
        return {k: v for k, v in self.__dict__.items() if k in sig.parameters}

    def _ensure_context(self, action):
        # An empty context means the action is relevant to the task-group.
        if len(action["context"]) == 0:
            return True

        for potential_matcher in action["context"]:
            if all(
                self.tags.get(key) == value for key, value in potential_matcher.items()
            ):
                return True

        return False

    def _get_action(self, decision_task, action_name):
        actions = decision_task.get_artifact("public/actions.json")
        for action in actions["actions"]:
            if action["name"] == action_name and self._ensure_context(action):
                assert action["kind"] == "hook"
                return action

        return None

    def _trigger_action(self, action, payload):
        tc_firefox_ci_credentials = config.get("taskcluster_firefox_ci", {})
        client_id = tc_firefox_ci_credentials.get("client_id")
        access_token = tc_firefox_ci_credentials.get("access_token")
        assert client_id and access_token, (
            "Missing Taskcluster Firefox CI credentials in mozci config secret"
        )

        options = taskcluster.optionsFromEnvironment()
        options["rootUrl"] = PRODUCTION_TASKCLUSTER_ROOT_URL
        options["credentials"] = {
            "clientId": client_id,
            "accessToken": access_token,
        }
        hooks = taskcluster.Hooks(options)

        result = hooks.triggerHook(action["hookGroupId"], action["hookId"], payload)
        return result["status"]["taskId"]

    def retrigger(self, push, times=3):
        """This function implements ability to perform retriggers on tasks."""
        decision_task = push.decision_task
        retrigger_action = self._get_action(decision_task, "retrigger-multiple")
        if not retrigger_action:
            logger.info(
                "Not retriggering task '{}', task should not be retriggered".format(
                    self.label
                )
            )
            return None

        hook_payload = jsone.render(
            retrigger_action["hookPayload"],
            context={
                "taskId": None,
                "taskGroupId": decision_task.id,
                "input": {"requests": [{"tasks": [self.label], "times": times}]},
            },
        )

        logger.info("Retriggering task '{}'".format(self.tags.get("label", "")))
        return self._trigger_action(retrigger_action, hook_payload)

    def confirm(self, push):
        """This function implements ability to trigger failure confirmation on tasks"""
        decision_task = push.decision_task
        retrigger_action = self._get_action(decision_task, "confirm-failures")
        if not retrigger_action:
            logger.info(
                "Not confirming task '{}', task should not be confirmed".format(
                    self.label
                )
            )
            return None

        hook_payload = jsone.render(
            retrigger_action["hookPayload"],
            context={
                "taskId": self.id,
                "taskGroupId": decision_task.id,
                "input": {},
            },
        )

        logger.info(
            "confirming failures for task '{}'".format(self.tags.get("label", ""))
        )
        return self._trigger_action(retrigger_action, hook_payload)

    def backfill(self, push, times: int | None = None):
        """
        This function implements ability to perform backfills on tasks.
        The number of task per push for the backfill can be set with the `times` argument.
        If unset, it is automatically detected depending on the classification of the task.
        """
        decision_task = push.decision_task
        backfill_action = self._get_action(decision_task, "backfill")

        if times is None:
            if (
                self.classification in ("not classified", "new failure not classified")
                or self.classification in INTERMITTENT_CLASSES
            ):
                times = 5
            else:
                times = 1
        hook_payload = jsone.render(
            backfill_action["hookPayload"],
            context={
                "taskId": self.id,
                "taskGroupId": decision_task.id,
                "input": {
                    "times": times
                    if self.classification
                    in ("not classified", "new failure not classified")
                    or self.classification in INTERMITTENT_CLASSES
                    else 1
                },
            },
        )

        logger.info("Backfilling task '{}'".format(self.tags.get("label", "")))
        return self._trigger_action(backfill_action, hook_payload)

    def is_build_failure(self) -> bool:
        """Returns whether a build task has failed."""
        return self.job_kind == "build" and self.failed

    def is_test_failure(self) -> bool:
        return self.job_kind == "test" and self.failed

    def can_retrigger_test_failure(self) -> bool:
        """Only allow to retrigger test tasks in failure state, with tier 1 or 2.
        Already classified tasks should not be retrigerred.
        If there is no task artifact which contains errorsummary in its name or is incomplete, then the task
        either hit an infrastructure issue or the test suite does not support failure summaries by errorsummary.
        """
        # TODO Use Task.is_confirmed_failure when it is ready
        if not self.is_test_failure():
            return False
        if self.tier not in (1, 2):
            return False
        if self.classification != "not classified":
            return False
        # Ignore backfills and retriggers, as only the initial failure should be triggered to avoid a loop.
        if self.is_backfill:
            return False
        if self.is_retrigger:
            return False
        if not any("errorsummary" in artifact for artifact in self.artifacts):
            logger.warning(
                f"Skipping retriggers for task {self.label}: task has no errorsummary artifact."
            )
            return False
        return True


@dataclass
class GroupResult:
    """Contains information relating to a single group failure within a TestTask."""

    group: str
    ok: bool
    # TODO: 'Optional' can be removed once https://github.com/mozilla/mozci/issues/662 is fixed.
    duration: Optional[int]


class FailureType(Enum):
    TIMEOUT = "timeout"
    CRASH = "crash"
    GENERIC = "generic"


TestName = NewType("TestName", str)
GroupName = NewType("GroupName", str)


@dataclass
class TestTask(Task):
    """Subclass containing additional information only relevant to 'test' tasks."""

    _results: Optional[List[GroupResult]] = field(default=None)
    _errors: Optional[List] = field(default=None)
    _failure_types: Optional[Dict[GroupName, List[Tuple[TestName, FailureType]]]] = (
        field(default=None)
    )

    @property
    def is_wpt(self):
        return any(
            s in self.label
            for s in {"web-platform-tests", "test-verify-wpt", "test-coverage-wpt"}
        )

    def retrieve_results(self, push):
        global slash_group_warned

        if is_no_groups_suite(self.label):
            assert self._errors is None, (
                f"{self.id} : {self.label} should have no errors"
            )
            self._errors = []

            assert self._results is None, (
                f"{self.id} : {self.label} should have no results"
            )
            self._results = []

            return

        if self.state == "completed":
            self._results = [
                GroupResult(group, result, duration)
                for group, (result, duration) in data.handler.get(
                    "test_task_groups", branch=push.branch, rev=push.rev, task=self
                ).items()
            ]
        else:
            self._results = []

        # Apply WPT workaround, needed at least until bug 1632546 is fixed.
        if self.is_wpt:
            # TODO: It can be removed a year after https://bugzilla.mozilla.org/show_bug.cgi?id=1688043 is fixed.
            # Filter out "/" groups.
            if not slash_group_warned and any(
                result.group == "/" for result in self._results
            ):
                slash_group_warned = True
                logger.warning(f"'/' group name in task {self.id}")

            self._results = [result for result in self._results if result.group != "/"]

            for result in self._results:
                result.group = wpt_workaround(result.group)

        # Filter out groups with bad names.
        # TODO: Figure out why we still have some groups with bad names.
        self._results = [
            result
            for result in self._results
            if not is_bad_group(self.id, result.group)
        ]

    @property
    def groups(self):
        return [result.group for result in self.results]

    @property
    def results(self):
        assert self._results is not None
        return self._results

    @property
    def errors(self):
        if self._errors is None:
            self._errors = data.handler.get("test_task_errors", task=self)
        return self._errors

    @property
    def failure_types(self):
        """
        Returns a dict mapping each failing group on this TestTask
        to a list of its failing test names and their FailureType.

        e.g:
        {"group/failing/on-this-task.ini": [
            ("group/failing/test-file-1.js", "timeout"),
            ("group/failing/test-file-2.js", "crash"),
        ]}
        """
        if self._failure_types is None:
            self._failure_types = data.handler.get(
                "test_task_failure_types", task_id=self.id
            )
        return self._failure_types


# Don't perform type checking because of https://github.com/python/mypy/issues/5374.
@dataclass  # type: ignore
class RunnableSummary(ABC):
    @property
    def is_intermittent(self):
        return self.status == Status.INTERMITTENT or any(
            c in INTERMITTENT_CLASSES for c, n in self.classifications
        )

    @property
    @abstractmethod
    def classifications(self): ...

    @property
    @abstractmethod
    def status(self): ...

    @property
    @abstractmethod
    def durations(self): ...

    @property
    @abstractmethod
    def total_duration(self): ...

    @property
    @abstractmethod
    def median_duration(self): ...


@dataclass
class GroupSummary(RunnableSummary):
    """Summarizes the overall state of a group (across retriggers)."""

    name: str
    tasks: List[TestTask]

    def __post_init__(self):
        # WPT names are not normalized relative to topsrcdir, so subsequent check
        # will fail unless normalized.
        if self.name.startswith("/"):
            self.name = wpt_workaround(self.name)
        assert all(self.name in t.groups for t in self.tasks)

    @property
    def classifications(self):
        return [
            (t.classification, t.classification_note)
            for t in self.tasks
            if t.failed
            and any(result.group == self.name and not result.ok for result in t.results)
        ]

    # TODO: Only used for testing, find a way to monkey patch!
    @classifications.setter
    def classifications(self, classifications):
        for c, t in zip(classifications, self.tasks):
            t.classification = c

    @property
    def durations(self) -> List[int]:
        data = []
        for task in self.tasks:
            for result in task.results:
                if result.group == self.name:
                    data.append(result.duration)
        return data

    @property
    def total_duration(self):
        return sum(self.durations)

    @property
    def median_duration(self):
        return median(self.durations)

    @memoized_property
    def status(self):
        overall_status_by_label = {}
        for task in self.tasks:
            for result in task.results:
                if result.group != self.name:
                    continue

                if not result.ok:
                    status = Status.FAIL
                else:
                    status = Status.PASS

                if task.label not in overall_status_by_label:
                    overall_status_by_label[task.label] = status
                elif status != overall_status_by_label[task.label]:
                    overall_status_by_label[task.label] = Status.INTERMITTENT

        # If the manifest failed intermittently at least in one task, we
        # consider it to be intermittent.
        if any(
            status == Status.INTERMITTENT for status in overall_status_by_label.values()
        ):
            return Status.INTERMITTENT

        # Otherwise, if the manifest failed at least once in any of the tasks,
        # we consider it as a failure.
        if any(status == Status.FAIL for status in overall_status_by_label.values()):
            return Status.FAIL

        # Otherwise, the manifest passed in all tasks, so we consider it a pass.
        return Status.PASS

    @memoized_property
    def failing_tasks(self):
        # List all tasks with some test results failing for that group
        return [
            task
            for task in self.tasks
            if any(
                not result.ok and result.group == self.name for result in task.results
            )
        ]

    def is_confirmed_failure(self) -> Optional[bool]:
        # match labels to -cf labels, and if ran, return True || False
        # if not run -cf task, return None
        # return False = intermittent
        # return True = regression
        cf_to_results = []

        for task in self.tasks:
            for result in task.results:
                if (
                    task.label
                    and task.label.endswith("-cf")
                    and result.group == self.name
                ):
                    cf_to_results.append(result.ok)

        if not cf_to_results:
            return None

        if False in cf_to_results:
            return True
        return False

    def is_config_consistent_failure(self, minimum_count: int = 3) -> Optional[bool]:
        config_to_results = collections.defaultdict(list)
        for task in self.tasks:
            for result in task.results:
                if result.group == self.name:
                    config_to_results[task.configuration].append(result.ok)

        # If there is no config for which we have at least 'minimum_count' runs, return None (that is, unknown).
        if all(len(results) < minimum_count for results in config_to_results.values()):
            return None

        # Return True if there is at least one configuration for which we have only failures, False otherwise.
        return any(
            len(results) >= minimum_count and not any(results)
            for results in config_to_results.values()
        )

    def is_cross_config_failure(self, minimum_count: int = 2) -> Optional[bool]:
        states = [
            result.ok
            for task in self.tasks
            for result in task.results
            if result.group == self.name
        ]

        nb = len(states)
        nb_passed = sum(states)  # Number of True booleans in the states list
        nb_failed = nb - nb_passed

        # If the group run on fewer than 'minimum_count' tasks, we don't have enough information to tell.
        if nb < minimum_count:
            return None

        # A group is a cross config failure when it is failing in all tasks and not
        # only in some.
        return nb_failed > 0 and nb_passed == 0

    # A group that is either failing across configurations or multiple times on the same configuration.
    def is_consistent_failure(
        self,
        cross_config_count: Optional[int] = 2,
        config_consistent_count: Optional[int] = 3,
    ) -> Optional[bool]:
        assert cross_config_count is not None or config_consistent_count is not None

        is_cross_config = None
        if cross_config_count is not None:
            is_cross_config = self.is_cross_config_failure(cross_config_count)
            if is_cross_config:
                return True

        is_config_consistent = None
        if config_consistent_count is not None:
            is_config_consistent = self.is_config_consistent_failure(
                config_consistent_count
            )
            if is_config_consistent:
                return True

        # We don't have enough information to tell if this is cross-config or consistently failing on the
        # same configuration (e.g. if we only run the test once on a single configuration, we can't tell
        # if it's cross-config or not and we can't tell if a retrigger would have failed).
        if is_cross_config is None and is_config_consistent is None:
            return None

        # If we got here, it means the group is:
        # - either failing on some configuration and passing on others;
        # - or failing on a run on a given configuration and passing on another run on the same configuration.
        return False


@dataclass
class LabelSummary(RunnableSummary):
    """Summarizes the overall state of a task label (across retriggers)."""

    label: str
    tasks: List[Task]

    def __post_init__(self):
        assert all(t.label == self.label for t in self.tasks)

    @property
    def classifications(self):
        return [
            (t.classification, t.classification_note) for t in self.tasks if t.failed
        ]

    @property
    def durations(self):
        return [t.duration for t in self.tasks]

    @property
    def total_duration(self):
        return sum(self.durations)

    @property
    def median_duration(self):
        return median(self.durations)

    @memoized_property
    def status(self):
        overall_status = None
        for task in self.tasks:
            if task.failed:
                status = Status.FAIL
            else:
                status = Status.PASS

            if overall_status is None:
                overall_status = status
            elif status != overall_status:
                overall_status = Status.INTERMITTENT

        return overall_status



# -*- coding: utf-8 -*-
import collections
import csv
import datetime
import fnmatch
import itertools
import json
import os
import re
import traceback
from abc import abstractmethod
from inspect import signature
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import urlencode

import arrow
import taskcluster
from cleo.commands.command import Command
from cleo.exceptions import CleoNoSuchOptionError
from cleo.helpers import argument, option
from loguru import logger
from tabulate import tabulate
from taskcluster.exceptions import TaskclusterRestFailure

from mozci import config
from mozci.errors import PushNotFound, SourcesNotFound, TaskNotFound
from mozci.push import (
    MAX_DEPTH,
    Push,
    PushStatus,
    Regressions,
    ToRetriggerOrBackfill,
    make_push_objects,
)
from mozci.task import Task, TestTask, is_autoclassifiable
from mozci.util.defs import INTERMITTENT_CLASSES
from mozci.util.hgmo import HgRev
from mozci.util.memoize import memoized_property
from mozci.util.taskcluster import (
    COMMUNITY_TASKCLUSTER_ROOT_URL,
    get_taskcluster_options,
    notify_email,
    notify_matrix,
)

EMAIL_CLASSIFY_EVAL = """
# classify-eval report generated on the {today}

The report contains statistics about pushes that were classified by Mozci.

## Statistics for the {total} pushes that were evaluated

{error_line}

{stats}
"""

EMAIL_PUSH_EVOLUTION = """
# Push {push.id} evolved from {previous} to {current}

Rev: [{push.rev}](https://treeherder.mozilla.org/jobs?repo={branch}&revision={push.rev})\n
Author: {push.author}\n
Time: {date}

## Real failures

- {real_failures}

"""

TWO_INTS_TUPLE_REGEXP = r"^\((\d+), ?(\d+)\)$"


class PushTasksCommand(Command):
    name = "push tasks"
    description = "List the tasks that ran on a push"
    arguments = [
        argument("rev", description="Head revision of the push."),
        argument(
            "branch", description="Branch the push belongs to (e.g autoland, try, etc)."
        ),
    ]

    def handle(self):
        push = Push(self.argument("rev"), self.argument("branch"))

        table = []
        for task in sorted(push.tasks, key=lambda t: t.label):
            table.append([task.label, task.result or "running"])

        self.line(tabulate(table, headers=["Label", "Result"]))


def check_type(parameter_type, hint, value):
    try:
        if parameter_type is bool:
            parameter = value not in [False, 0, "0", "False", "false", "f"]
        elif parameter_type == Optional[Tuple[int, int]]:
            match = re.match(TWO_INTS_TUPLE_REGEXP, value)

            if not match or len(match.groups()) != 2:
                raise ValueError

            parameter = tuple([int(number) for number in match.groups()])
        else:
            parameter = parameter_type(value)
    except ValueError:
        raise Exception(
            f"Provided {hint} should be a {parameter_type.__name__ if hasattr(parameter_type, '__name__') else parameter_type}."
        )

    return parameter


def retrieve_classify_parameters(options):
    default_parameters = []
    for name, parameter in signature(Push.classify).parameters.items():
        if name != "self":
            default_parameters.append(parameter)

    classify_parameters = {}
    for parameter in default_parameters:
        parameter_name = parameter.name
        parameter_type = parameter.annotation
        option_name = parameter_name.replace("_", "-")
        try:
            option = options(option_name)
        except CleoNoSuchOptionError:
            option = None

        if config.get(parameter_name) is not None and option is not None:
            raise Exception(
                f"You should provide either --{option_name} CLI option OR '{parameter_name}' in the config secret not both."
            )

        if config.get(parameter_name) is not None:
            classify_parameters[parameter_name] = check_type(
                parameter_type,
                f"'{parameter_name}' in the config secret",
                config[parameter_name],
            )
        elif option is not None:
            classify_parameters[parameter_name] = check_type(
                parameter_type, f"--{option_name}", option
            )

    return classify_parameters


class BasePushCommand(Command):
    """
    Provide abstraction class to fetch a list of pushes based on provided arguments.
    """

    arguments = [
        argument(
            "branch",
            description="Branch the push belongs to (e.g autoland, try, etc).",
            optional=True,
            default="autoland",
        )
    ]
    options = [
        option("rev", description="Head revision of the push.", flag=False),
        option(
            "from-date",
            description='Lower bound of the push range (as a date in yyyy-mm-dd format or a human expression like "1 days ago").',
            flag=False,
        ),
        option(
            "to-date",
            description='Upper bound of the push range (as a date in yyyy-mm-dd format or a human expression like "1 days ago"), defaults to now.',
            flag=False,
        ),
    ]

    def list_pushes(
        self, branch: str, from_date: str, to_date: str, rev: str
    ) -> List[Push]:
        if not (bool(rev) ^ bool(from_date or to_date)):
            raise Exception(
                "You must either provide a single push revision with --rev or define at least --from-date option to classify a range of pushes (note: --to-date will default to current time if not given)."
            )

        if rev:
            pushes = [Push(rev, branch)]
        else:
            if not from_date:
                raise Exception(
                    "You must provide at least --from-date to classify a range of pushes (note: --to-date will default to current time if not given)."
                )

            now = datetime.datetime.now()
            if not to_date:
                to_date = datetime.datetime.strftime(now, "%Y-%m-%d")

            arrow_now = arrow.get(now)
            try:
                datetime.datetime.strptime(from_date, "%Y-%m-%d")
            except ValueError:
                try:
                    from_date = arrow_now.dehumanize(from_date).format("YYYY-MM-DD")
                except ValueError:
                    raise Exception(
                        'Provided --from-date should be a date in yyyy-mm-dd format or a human expression like "1 days ago".'
                    )

            try:
                datetime.datetime.strptime(to_date, "%Y-%m-%d")
            except ValueError:
                try:
                    to_date = arrow_now.dehumanize(to_date).format("YYYY-MM-DD")
                except ValueError:
                    raise Exception(
                        'Provided --to-date should be a date in yyyy-mm-dd format or a human expression like "1 days ago".'
                    )

            pushes = make_push_objects(
                from_date=from_date, to_date=to_date, branch=branch
            )

        return pushes

    @property
    def branch(self):
        return self.argument("branch")

    @memoized_property
    def pushes(self) -> list[Push]:
        return self.list_pushes(
            self.branch,
            self.option("from-date"),
            self.option("to-date"),
            self.option("rev"),
        )

    @abstractmethod
    def handle(self): ...


class ClassifyCommand(BasePushCommand):
    name = "push classify"
    description = "Display the classification for a given push (or a range of pushes) as GOOD, BAD or UNKNOWN"

    options = BasePushCommand.options + [
        option(
            "intermittent-confidence-threshold",
            description="Medium confidence threshold used to classify the regressions.",
            flag=False,
        ),
        option(
            "real-confidence-threshold",
            description="High confidence threshold used to classify the regressions.",
            flag=False,
        ),
        option(
            "use-possible-regressions",
            description="Use possible regressions while classifying the regressions.",
            default=None,
            flag=False,
        ),
        option(
            "unknown-from-regressions",
            description="Unknown from regressions while classifying the regressions.",
            default=None,
            flag=False,
        ),
        option(
            "consider-children-pushes-configs",
            description="Consider children pushes configs while classifying the regressions.",
            default=None,
            flag=False,
        ),
        option(
            "cross-config-counts",
            description="Cross-config counts used to classify the regressions.",
            flag=False,
        ),
        option(
            "consistent-failures-counts",
            description="Consistent failures counts used to classify the regressions.",
            flag=False,
        ),
        option(
            "output",
            description="Path towards a directory to save a JSON file containing classification and regressions details in.",
            flag=False,
        ),
        option(
            "show-intermittents",
            description="If set, print tasks that should be marked as intermittent.",
            flag=True,
        ),
        option(
            "environment",
            description="Environment in which the analysis is running (testing, production, ...).",
            flag=False,
            default="testing",
        ),
        option(
            "retrigger-limit",
            description="Number of failing groups (missing information) to be retriggered.",
            flag=False,
            default=3,
        ),
        option(
            "backfill-limit",
            description="Number of failing groups (missing information) to be backfilled.",
            flag=False,
            default=3,
        ),
        option(
            "disable-build-regressions",
            description="If set, skip the analysis of build tasks regressions.",
            flag=True,
        ),
        option(
            "maximum_build_tasks_to_retrigger",
            description=(
                "The maximum number of build task failures to retrigger for a single push. "
                "The value can be set to -1 for no maximum."
            ),
            flag=False,
            default=15,
        ),
    ]

    def handle(self) -> None:
        classify_parameters = retrieve_classify_parameters(self.option)

        output = self.option("output")
        if output and not os.path.isdir(output):
            os.makedirs(output)
            self.line(
                "<comment>Provided --output pointed to a inexistent directory that is now created.</comment>"
            )

        retriggerable_backfillable_patterns = config.get(
            "retriggerable-backfillable-task-names", []
        )

        try:
            retrigger_limit = int(self.option("retrigger-limit"))
        except ValueError:
            raise Exception("Provided --retrigger-limit should be an int.")

        try:
            backfill_limit = int(self.option("backfill-limit"))
        except ValueError:
            raise Exception("Provided --backfill-limit should be an int.")

        for push in self.pushes:
            if self.option("disable-build-regressions"):
                self.line("Skipping build regressions analysis")
            else:
                try:
                    tasks_to_retrigger = push.check_build_regressions()
                    if not tasks_to_retrigger:
                        self.line("No build task detected as potential regression")
                    else:
                        max_retrigger = self.option("maximum_build_tasks_to_retrigger")
                        if max_retrigger > 0:
                            if len(tasks_to_retrigger) > max_retrigger:
                                # In case many build failures are observed, the issue is likely not
                                # general intermittentness of the tasks if the tree/repository/project
                                # is mozilla-central but either a permanent failure of the executed
                                # task payload or an infrastructure, e.g. to access required resources
                                self.line(
                                    f"<error> More than {max_retrigger} build tasks were detected as potential regressions. "
                                    "Only the first {max_retrigger} will be retriggered."
                                )
                            tasks_to_retrigger = tasks_to_retrigger[:max_retrigger]
                        self.line(
                            f"Retriggering {len(tasks_to_retrigger)} build regression that may introduce a build bustage"
                        )
                        for task in tasks_to_retrigger:
                            task.retrigger(push)
                except Exception as e:
                    self.line(
                        f"<error> Couldn't run build regressions analysis on push {push.push_uuid}: {e}.</error>"
                    )

            try:
                classification, regressions, to_retrigger_or_backfill = push.classify(
                    **classify_parameters
                )

                self.backfill_and_retrigger_failures(
                    push,
                    retriggerable_backfillable_patterns,
                    classify_parameters,
                    retrigger_limit,
                    backfill_limit,
                    to_retrigger_or_backfill,
                )

                self.line(
                    f"Push associated with the head revision {push.rev} on "
                    f"the branch {self.branch} is classified as {classification.name}"
                )
            except Exception as e:
                self.line(
                    f"<error>Couldn't classify push {push.push_uuid}: {e}.</error>"
                )
                # Print the error stacktrace in red
                self.line(f"<error>{traceback.format_exc()}</error>")
                continue

            # Rettriger test failures to detect permanent ones, only on Linux cloud worker pool for now
            push.identify_permanent_failures(queue_prefix="gecko-t/t-linux-")

            if self.option("show-intermittents"):
                self.line("-" * 50)
                self.line(
                    "Printing tasks that should be marked as intermittent failures:"
                )
                for task in regressions.intermittent:
                    self.line(task)
                self.line("-" * 50)

            if output:

                def _serialize_regressions(regressions):
                    return {
                        group: [
                            {
                                "task_id": task.id,
                                "label": task.label,
                                "autoclassify": is_autoclassifiable(task),
                                "tests": [
                                    test_name
                                    for group_failures in task.failure_types.values()
                                    for test_name, _ in group_failures
                                ],
                            }
                            for task in failing_tasks
                        ]
                        for group, failing_tasks in regressions.items()
                    }

                to_save = {
                    "push": {
                        "id": push.push_uuid,
                        "classification": classification.name,
                    },
                    "failures": {
                        "real": _serialize_regressions(regressions.real),
                        "intermittent": _serialize_regressions(
                            regressions.intermittent
                        ),
                        "unknown": _serialize_regressions(regressions.unknown),
                    },
                }

                filename = f"{output}/classify_output_{self.branch}_{push.rev}.json"
                with open(filename, "w") as file:
                    json.dump(to_save, file, indent=2)

                self.line(
                    f"Classification and regressions details for push {push.push_uuid} were saved in {filename} JSON file"
                )

            # Send a notification when some emails are declared in the config
            emails = config.get("emails", {}).get("classifications")
            matrix_room = config.get("matrix-room-id")
            if emails or matrix_room:
                # Load previous classification from taskcluster
                try:
                    previous = push.get_existing_classification(
                        self.option("environment")
                    )
                except SourcesNotFound:
                    # We still want to send a notification if the current one is bad
                    previous = None

                self.send_notifications(
                    emails, matrix_room, push, previous, classification, regressions
                )

    def retrigger_failures(
        self, push, groups, count, allowed_patterns, retrigger_limit
    ):
        groups_with_failures = {}
        for name, failing_tasks in groups.items():
            filtered_failing_tasks = [
                task
                for task in failing_tasks
                if any(
                    fnmatch.fnmatch(task.label, pattern) for pattern in allowed_patterns
                )
            ]
            if filtered_failing_tasks:
                assert all(
                    any(
                        not result.ok and result.group == name
                        for result in task.results
                    )
                    for task in filtered_failing_tasks
                ), (
                    f"Some failing tasks on the group {name} (to be retriggered) didn't really fail"
                )
                groups_with_failures[name] = filtered_failing_tasks

        if not groups_with_failures:
            return

        for failing_tasks in itertools.islice(
            groups_with_failures.values(), 0, retrigger_limit
        ):
            # If there is more than one task failing in this group, we should retrigger only one of them
            failing_tasks[0].confirm(push)

    def backfill_and_retrigger_failures(
        self,
        push: Push,
        allowed_patterns: List[str],
        classify_parameters: Dict[str, Any],
        retrigger_limit: int,
        backfill_limit: int,
        to_retrigger_or_backfill: ToRetriggerOrBackfill,
    ) -> None:
        # Retrigger real failures
        # TODO: Potential real failures might be coming from children pushes too, but we should instead
        # only retrigger tasks on the push where they were defined (https://github.com/mozilla/mozci/issues/796).
        self.line("Retriggering failures to ensure they are real")
        self.retrigger_failures(
            push,
            to_retrigger_or_backfill.real_retrigger,
            classify_parameters.get("consistent_failures_counts", (2, 3))[1],
            allowed_patterns,
            retrigger_limit,
        )

        # Retrigger intermittent failures
        self.line("Retriggering failures to ensure they are intermittent")
        self.retrigger_failures(
            push,
            to_retrigger_or_backfill.intermittent_retrigger,
            classify_parameters.get("consistent_failures_counts", (2, 3))[0],
            allowed_patterns,
            retrigger_limit,
        )

        # Backfill some failures
        self.line("Backfilling failures to ensure they are caused by this push")
        groups_to_backfill = {
            name: failing_tasks
            for name, failing_tasks in to_retrigger_or_backfill.backfill.items()
            if failing_tasks
        }
        for failing_tasks in itertools.islice(
            groups_to_backfill.values(), 0, backfill_limit
        ):
            for t in failing_tasks:
                if t.label and any(
                    fnmatch.fnmatch(t.label, pattern) for pattern in allowed_patterns
                ):
                    t.backfill(push)

    def send_notifications(
        self,
        emails: Optional[List[str]],
        matrix_room: Optional[str],
        push: Push,
        previous: Optional[PushStatus],
        current: PushStatus,
        regressions: Regressions,
    ) -> None:
        """
        Send an email and/or a matrix notification when:
        - there is no previous classification and the new classification is BAD;
        - the previous classification was GOOD or UNKNOWN and the new classification is BAD;
        - or the previous classification was BAD and the new classification is GOOD or UNKNOWN.
        """

        def _get_task_url(task: TestTask):
            """Helper to build a treeherder link for a task"""
            assert task.id is not None
            params = {
                "repo": self.branch,
                "revision": push.rev,
                "selectedTaskRun": f"{task.id}-0",
            }

            return f"https://treeherder.mozilla.org/#/jobs?{urlencode(params)}"

        def _get_group_url(group: str):
            """Helper to build a treeherder link for a group"""
            params = {"repo": self.branch, "tochange": push.rev, "test_paths": group}
            return f"https://treeherder.mozilla.org/#/jobs?{urlencode(params)}"

        def _list_tasks(tasks):
            """Helper to build a list of all tasks in a group, with their treeherder url"""
            if not tasks:
                return "No tasks available"

            return "Tasks:\n  - " + "\n  - ".join(
                [f"[{task.label}]({_get_task_url(task)})" for task in tasks]
            )

        if (
            previous in (None, PushStatus.GOOD, PushStatus.UNKNOWN)
            and current == PushStatus.BAD
        ) or (
            previous == PushStatus.BAD
            and current in (PushStatus.GOOD, PushStatus.UNKNOWN)
        ):
            formatted_date = datetime.datetime.fromtimestamp(
                push.date, tz=datetime.timezone.utc
            ).strftime("%H:%M:%S")
            email_content = EMAIL_PUSH_EVOLUTION.format(
                previous=previous.name if previous else "no classification",
                current=current.name,
                push=push,
                date=formatted_date,
                branch=self.branch,
                real_failures="\n- ".join(
                    [
                        f"Group [{group}]({_get_group_url(group)}) - {_list_tasks(tasks)}"
                        for group, tasks in regressions.real.items()
                    ]
                ),
            )
            if emails:
                notify_email(
                    emails=emails,
                    subject=f"Push status evolution {push.id} {push.rev[:8]}",
                    content=email_content,
                )
            if matrix_room:
                notify_matrix(
                    room=matrix_room,
                    body=email_content,
                )


def prepare_for_analysis(push):
    removed_tasks: Dict[str, List[Task]] = {}
    backedoutby: Dict[str, str] = {}
    old_classifications: Dict[str, Dict[str, Dict[str, str]]] = {}

    all_pushes = set(
        [push]
        + [parent for parent in push._iterate_parents(max_depth=MAX_DEPTH)]
        + [child for child in push._iterate_children(max_depth=MAX_DEPTH)]
    )
    for p in all_pushes:
        # Ignore retriggers and backfills on current push/its parents/its children.
        removed_tasks[p.id] = [
            task for task in p.tasks if task.is_backfill or task.is_retrigger
        ]
        p.tasks = [task for task in p.tasks if task not in removed_tasks[p.id]]

        # Pretend push was not backed out.
        backedoutby[p.id] = p.backedoutby
        p.backedoutby = None

        # Pretend push was not finalized yet.
        p._date = datetime.datetime.now().timestamp()

        # Pretend no tasks were classified to run the model without any outside help.
        old_classifications[p.id] = {}
        for task in p.tasks:
            old_classifications[p.id][task.id] = {
                "classification": task.classification,
                "note": task.classification_note,
            }
            task.classification = "not classified"
            task.classification_note = None

    return all_pushes, removed_tasks, backedoutby, old_classifications


def retrieve_sheriff_reals(pushes_group_summaries, push):
    # Compare real failures that were predicted by mozci with the ones classified by Sheriffs
    sheriff_reals = set()
    # Get likely regressions of this push
    likely_regressions = push.get_likely_regressions("group", True)
    # Only consider groups that were classified as "fixed by commit" to exclude likely regressions mozci found via heuristics.
    max_depth = None if push.backedout or push.bustage_fixed_by else MAX_DEPTH
    for other in push._iterate_children(max_depth=max_depth):
        if other.push_uuid not in pushes_group_summaries:
            pushes_group_summaries[other.push_uuid] = other.group_summaries

        for name, group in pushes_group_summaries[other.push_uuid].items():
            classifications = set([c for c, _ in group.classifications])
            if classifications == {"fixed by commit"} and name in likely_regressions:
                sheriff_reals.add(name)

    return pushes_group_summaries, sheriff_reals


def retrieve_sheriff_intermittents(pushes_group_summaries, push):
    if push.push_uuid not in pushes_group_summaries:
        pushes_group_summaries[push.push_uuid] = push.group_summaries

    # Compare intermittent failures that were predicted by mozci with the ones classified by Sheriffs
    sheriff_intermittents = set()
    for name, group in pushes_group_summaries[push.push_uuid].items():
        classifications = set([c for c, _ in group.classifications])
        if classifications <= set(INTERMITTENT_CLASSES):
            sheriff_intermittents.add(name)

    return pushes_group_summaries, sheriff_intermittents


def parse_and_log_details(
    group_summaries,
    sheriff_groups,
    expected,
    push=None,
    failures=None,
    predicted_groups=None,
    ignore_pending_conflicting=False,
    state="",
    suffix="",
):
    to_print = []

    if failures and not predicted_groups:
        predicted_groups = failures[push][state].keys() if failures.get(push) else []

    total = len(predicted_groups)
    if not total:
        if sheriff_groups:
            to_print.append(
                f"{len(sheriff_groups)} groups were classified as {state} by Sheriffs and missed by Mozci, missed groups:"
            )
            to_print.append("  - " + "\n  - ".join(sheriff_groups))

        output = {
            f"total{suffix}": 0,
            f"correct{suffix}": 0,
            f"wrong{suffix}": 0,
            f"missed{suffix}": len(sheriff_groups),
        }
        if not ignore_pending_conflicting:
            output[f"pending{suffix}"] = 0
            output[f"conflicting{suffix}"] = 0

        return output, to_print

    conflicting = []
    differing = []
    pending = []
    for group in predicted_groups:
        classifications_set = set(
            task.classification for task in group_summaries[group].tasks if task.failed
        )
        if len(classifications_set) == 0:
            continue
        if (
            len(classifications_set - {"not classified", "new failure not classified"})
            == 0
        ):
            pending.append(group)
        elif len(classifications_set) != 1:
            conflicting.append(group)
        elif classifications_set.isdisjoint(expected):
            differing.append(group)

    missed = []
    for group in sheriff_groups:
        if group not in predicted_groups:
            missed.append(group)

    correct = total - len(differing)
    if not ignore_pending_conflicting:
        correct -= len(pending) + len(conflicting)

    to_print.append(
        f"{correct} out of {total} {state} groups were also classified as {state} by Sheriffs."
    )
    if differing:
        to_print.append(
            f"{len(differing)} out of {total} {state} groups weren't classified as {state} by Sheriffs, differing groups:"
        )
        to_print.append("  - " + "\n  - ".join(differing))
    if not ignore_pending_conflicting:
        if pending:
            to_print.append(
                f"{len(pending)} out of {total} {state} groups are waiting to be classified by Sheriffs."
            )
        if conflicting:
            to_print.append(
                f"{len(conflicting)} out of {total} {state} groups have conflicting classifications applied by Sheriffs, inconsistent groups:"
            )
            to_print.append("  - " + "\n  - ".join(conflicting))
    if missed:
        to_print.append(
            f"{len(missed)} groups were classified as {state} by Sheriffs and missed (or classified as unknown) by Mozci, missed groups:"
        )
        to_print.append("  - " + "\n  - ".join(missed))

    output = {
        f"total{suffix}": total,
        f"correct{suffix}": correct,
        f"wrong{suffix}": len(differing),
        f"missed{suffix}": len(missed),
    }
    if not ignore_pending_conflicting:
        output[f"pending{suffix}"] = len(pending)
        output[f"conflicting{suffix}"] = len(conflicting)

    return output, to_print


def check_ever_classified_as_cause(push, iterate_on):
    ever_classified_as_cause = False
    for (
        other,
        _,
        candidate_regressions,
        classified_as_cause,
    ) in push._iterate_failures(iterate_on):
        if push.backedoutby in other.revs or push.bustage_fixed_by in other.revs:
            return ever_classified_as_cause

        ever_classified_as_cause = any(
            result is True
            for name in candidate_regressions.keys()
            for result in classified_as_cause[name]
        )
        if ever_classified_as_cause:
            return ever_classified_as_cause


class ClassifyEvalCommand(Command):
    name = "push classify-eval"
    description = "Evaluate the classification results for a given push (or a range of pushes) by comparing them with reality"
    options = BasePushCommand.options + [
        option(
            "intermittent-confidence-threshold",
            description="Medium confidence threshold used to classify the regressions.",
            flag=False,
        ),
        option(
            "real-confidence-threshold",
            description="High confidence threshold used to classify the regressions.",
            flag=False,
        ),
        option(
            "use-possible-regressions",
            description="Use possible regressions while classifying the regressions.",
            default=None,
            flag=False,
        ),
        option(
            "unknown-from-regressions",
            description="Unknown from regressions while classifying the regressions.",
            default=None,
            flag=False,
        ),
        option(
            "consider-children-pushes-configs",
            description="Consider children pushes configs while classifying the regressions.",
            default=None,
            flag=False,
        ),
        option(
            "cross-config-counts",
            description="Cross-config counts used to classify the regressions.",
            flag=False,
        ),
        option(
            "consistent-failures-counts",
            description="Consistent failures counts used to classify the regressions.",
            flag=False,
        ),
        option(
            "recalculate",
            description="If set, recalculate the classification instead of fetching an artifact.",
            flag=True,
        ),
        option(
            "output",
            description="Path towards a path to save a CSV file with classification states for various pushes.",
            flag=False,
        ),
        option(
            "send-email",
            description="If set, also send the evaluation report by email instead of just logging it.",
            flag=True,
        ),
        option(
            "detailed-classifications",
            description="If set, compare real/intermittent group classifications with Sheriff's ones.",
            flag=True,
        ),
        option(
            "environment",
            description="Environment in which the analysis is running (testing, production, ...).",
            flag=False,
            default="testing",
        ),
    ]

    def handle(self) -> None:
        branch = self.argument("branch")

        self.line("<comment>Loading pushes...</comment>")

        option_names = [
            name.replace("_", "-")
            for name, _ in signature(Push.classify).parameters.items()
            if name != "self"
        ]
        if self.option("recalculate"):
            classify_parameters = retrieve_classify_parameters(self.option)
        elif any(self.option(name) for name in option_names):
            self.line(
                f"<error>--recalculate isn't set, you shouldn't provide --{', --'.join(option_names)} CLI options.</error>"
            )
            return

        # Progress bar will display time stats & messages
        progress = self.progress_bar(len(self.pushes))
        progress.set_format(
            " %current%/%max% [%bar%] %percent:3s%% %elapsed:6s% %message%"
        )

        # Setup specific route prefix for existing tasks, according to environment
        environment = self.option("environment")
        route_prefix = (
            "project.mozci.classification"
            if environment == "production"
            else f"project.mozci.{environment}.classification"
        )

        self.errors = {}
        self.classifications = {}
        self.failures = {}
        for push in self.pushes:
            if self.option("recalculate"):
                progress.set_message(f"Calc. {branch} {push.id}")

                (
                    all_pushes,
                    removed_tasks,
                    backedoutby,
                    old_classifications,
                ) = prepare_for_analysis(push)

                try:
                    self.classifications[push], regressions, _ = push.classify(
                        **classify_parameters
                    )
                    self.failures[push] = {
                        "real": regressions.real,
                        "intermittent": regressions.intermittent,
                        "unknown": regressions.unknown,
                    }
                except Exception as e:
                    self.line(
                        f"<error>Classification failed on {branch} {push.rev}: {e}</error>"
                    )
                    self.errors[push] = e

                for p in all_pushes:
                    # Once the Mozci algorithm has run, restore Sheriffs classifications to be able to properly compare failures classifications.
                    for task in p.tasks:
                        task.classification = old_classifications[p.id][task.id][
                            "classification"
                        ]
                        task.classification_note = old_classifications[p.id][task.id][
                            "note"
                        ]

                    # Restore backout information.
                    p.backedoutby = backedoutby[p.id]

                    # And also restore tasks marked as a backfill or a retrigger.
                    p.tasks = p.tasks + removed_tasks[p.id]
            else:
                progress.set_message(f"Fetch {branch} {push.id}")
                try:
                    index = f"{route_prefix}.{branch}.revision.{push.rev}"
                    task = Task.create(
                        index=index, root_url=COMMUNITY_TASKCLUSTER_ROOT_URL
                    )

                    artifact = task.get_artifact(
                        "public/classification.json",
                        root_url=COMMUNITY_TASKCLUSTER_ROOT_URL,
                    )
                    self.classifications[push] = PushStatus[
                        artifact["push"]["classification"]
                    ]
                    self.failures[push] = artifact["failures"]
                except TaskNotFound as e:
                    self.line(
                        f"<comment>Taskcluster task missing for {branch} {push.rev}</comment>"
                    )
                    self.errors[push] = e

                except Exception as e:
                    self.line(
                        f"<error>Fetch failed on {branch} {push.rev}: {e}</error>"
                    )
                    self.errors[push] = e

            warnings = []
            # Warn about pushes that are backed-out and where all failures on the push itself and its children are marked as intermittent
            if push.backedout or push.bustage_fixed_by:
                ever_classified_as_cause = check_ever_classified_as_cause(push, "label")

                if not ever_classified_as_cause:
                    ever_classified_as_cause = check_ever_classified_as_cause(
                        push, "group"
                    )

                if not ever_classified_as_cause:
                    warnings.append(
                        {
                            "message": f"Push {push.branch}/{push.rev} was backedout and all of its failures and the ones of its children were marked as intermittent or marked as caused by another push.",
                            "type": "error",
                            "notify": config.get("warnings", {}).get(
                                "ever_classified_as_cause", False
                            ),
                        }
                    )

            for task in push.tasks:
                if task.classification != "fixed by commit":
                    continue

                # Warn if there is a classification that references a revision that does not exist
                fix_hgmo = HgRev.create(
                    task.classification_note[:12], branch=push.branch
                )
                try:
                    fix_hgmo.changesets
                except PushNotFound:
                    warnings.append(
                        {
                            "message": f"Task {task.id} on push {push.branch}/{push.rev} contains a classification that references a non-existent revision: {task.classification_note}.",
                            "type": "error",
                            "notify": config.get("warnings", {}).get(
                                "non_existent_fix", False
                            ),
                        }
                    )
                    continue

                if fix_hgmo.pushid <= push.id:
                    warnings.append(
                        {
                            "message": f"Task {task.label} on push {push.branch}/{push.rev} is classified as fixed by {task.classification_note}, which is older than the push itself.",
                            "type": "error",
                            "notify": config.get("warnings", {}).get(
                                "fix_older_than_push", False
                            ),
                        }
                    )
                    continue

                # Warn when a failure is classified as fixed by a backout of a push that is newer than the failure itself
                all_backedouts = set(
                    backedout
                    for backedouts in fix_hgmo.backouts.values()
                    for backedout in backedouts
                )

                all_bustagefixed = set()
                for child in push._iterate_children():
                    if child.rev == fix_hgmo.node:
                        break

                    for bug in child.bugs:
                        if bug in fix_hgmo.bugs_without_backouts:
                            all_bustagefixed.add(child.rev)

                all_fixed = all_backedouts | all_bustagefixed

                if len(all_fixed) > 0 and all(
                    HgRev.create(backedout, branch=push.branch).pushid > push.id
                    for backedout in all_fixed
                ):
                    warnings.append(
                        {
                            "message": f"Task {task.label} on push {push.branch}/{push.rev} is classified as fixed by a backout/bustage fix ({fix_hgmo.node}) of pushes ({all_fixed}) that come after the failure itself.",
                            "type": "error",
                            "notify": config.get("warnings", {}).get(
                                "backout_of_newer_pushes", False
                            ),
                        }
                    )

            # Warn when there are inconsistent classifications for a given group
            if push.backedout or push.bustage_fixed_by:
                group_classifications: dict[str, dict[tuple[str, str], set[str]]] = (
                    collections.defaultdict(lambda: collections.defaultdict(set))
                )
                for other in push._iterate_children():
                    if (
                        push.backedoutby in other.revs
                        or push.bustage_fixed_by in other.revs
                    ):
                        break

                    for name, summary in other.group_summaries.items():
                        for classification in summary.classifications:
                            group_classifications[name][classification].add(other.rev)

                for name, classification_to_revs in group_classifications.items():
                    if len(classification_to_revs) > 1:
                        inconsistent_list = [
                            f"  - {classification} in pushes {', '.join(revs)}"
                            for classification, revs in classification_to_revs.items()
                        ]
                        inconsistent = "\n" + ",\n".join(inconsistent_list)
                        warnings.append(
                            {
                                "message": f"Group {name} has inconsistent classifications: {inconsistent}.",
                                "type": "comment",
                                "notify": config.get("warnings", {}).get(
                                    "inconsistent", False
                                ),
                            }
                        )

            # Output all warnings and also send them to the Matrix room if defined
            matrix_room = config.get("matrix-room-id")
            for warning in warnings:
                warn_type = warning["type"]
                warn_message = warning["message"]
                do_notify = warning["notify"]
                self.line(f"<{warn_type}>{warn_message}</{warn_type}>")
                if matrix_room and do_notify:
                    notify_matrix(room=matrix_room, body=warn_message)

            if not matrix_room and warnings:
                self.line(
                    "<comment>Some warning notifications should have been sent but no matrix room was provided in the secret.</comment>"
                )

            # Advance the overall progress bar
            progress.advance()

        # Conclude the progress bar
        progress.finish()
        print("\n")

        error_line = ""
        if self.errors:
            if self.option("recalculate"):
                error_line = "Failed to recalculate classification"
            else:
                error_line = "Failed to fetch classification"

            error_line += f" for {len(self.errors)} out of {len(self.pushes)} pushes."

            if not self.option("recalculate") and not self.option("send-email"):
                error_line += " Use the '--recalculate' option if you want to generate them yourself."

            self.line(f"<error>{error_line}</error>")

        stats = [
            self.log_pushes(PushStatus.BAD, False),
            self.log_pushes(PushStatus.BAD, True),
            self.log_pushes(PushStatus.GOOD, False),
            self.log_pushes(PushStatus.GOOD, True),
            self.log_pushes(PushStatus.UNKNOWN, False),
            self.log_pushes(PushStatus.UNKNOWN, True),
        ]

        if self.option("detailed-classifications"):
            self.line("\n")

            pushes_group_summaries = {}
            real_stats = intermittent_stats = {
                "total": 0,
                "correct": 0,
                "wrong": 0,
                "pending": 0,
                "conflicting": 0,
                "missed": 0,
            }
            for push in self.pushes:
                self.line(
                    f"<comment>Printing detailed classifications comparison for push {push.branch}/{push.rev}</comment>"
                )

                if push.push_uuid not in pushes_group_summaries:
                    pushes_group_summaries[push.push_uuid] = push.group_summaries

                # Compare real failures that were predicted by mozci with the ones classified by Sheriffs
                try:
                    pushes_group_summaries, sheriff_reals = retrieve_sheriff_reals(
                        pushes_group_summaries, push
                    )
                except Exception:
                    self.line(
                        "<error>Failed to retrieve Sheriff classifications for the real failures of this push.</error>"
                    )

                try:
                    push_real_stats, to_print = parse_and_log_details(
                        pushes_group_summaries[push.push_uuid],
                        sheriff_reals,
                        {"fixed by commit"},
                        push=push,
                        failures=self.failures,
                        state="real",
                    )

                    for line in to_print:
                        self.line(line)

                    real_stats = {
                        key: value + push_real_stats[key]
                        for key, value in real_stats.items()
                    }
                except Exception:
                    self.line(
                        "<error>Failed to compare true and predicted real failures of this push.</error>"
                    )

                # Compare intermittent failures that were predicted by mozci with the ones classified by Sheriffs
                try:
                    (
                        pushes_group_summaries,
                        sheriff_intermittents,
                    ) = retrieve_sheriff_intermittents(pushes_group_summaries, push)
                except Exception:
                    self.line(
                        "<error>Failed to retrieve Sheriff classifications for the intermittent failures of this push.</error>"
                    )

                try:
                    push_intermittent_stats, to_print = parse_and_log_details(
                        pushes_group_summaries[push.push_uuid],
                        sheriff_intermittents,
                        set(INTERMITTENT_CLASSES),
                        push=push,
                        failures=self.failures,
                        state="intermittent",
                    )

                    for line in to_print:
                        self.line(line)

                    intermittent_stats = {
                        key: value + push_intermittent_stats[key]
                        for key, value in intermittent_stats.items()
                    }
                except Exception:
                    self.line(
                        "<error>Failed to compare true and predicted intermittent failures of this push.</error>"
                    )

            self.line(
                f"\n<comment>Printing overall detailed classifications comparison for {len(self.pushes)} pushes</comment>"
            )
            detailed_stats = [
                f"{real_stats['correct']} out of {real_stats['total']} failures were correctly classified as real ('fixed by commit' by Sheriffs).",
                f"{real_stats['wrong']} out of {real_stats['total']} failures were wrongly classified as real ('intermittent' by Sheriffs).",
                f"{real_stats['pending']} out of {real_stats['total']} failures classified as real are waiting to be classified by Sheriffs.",
                f"{real_stats['conflicting']} out of {real_stats['total']} failures classified as real have conflicting classifications applied by Sheriffs.",
                f"{real_stats['missed']} real failures were missed or classified as unknown by Mozci.",
                f"{intermittent_stats['correct']} out of {intermittent_stats['total']} failures were correctly classified as intermittent ('intermittent' by Sheriffs).",
                f"{intermittent_stats['wrong']} out of {intermittent_stats['total']} failures were wrongly classified as intermittent ('fixed by commit' by Sheriffs).",
                f"{intermittent_stats['pending']} out of {intermittent_stats['total']} failures classified as intermittent are waiting to be classified by Sheriffs.",
                f"{intermittent_stats['conflicting']} out of {intermittent_stats['total']} failures classified as intermittent have conflicting classifications applied by Sheriffs.",
                f"{intermittent_stats['missed']} intermittent failures were missed or classified as unknown by Mozci.",
            ]
            for line in detailed_stats:
                self.line(line)

            stats += detailed_stats

        if self.option("send-email"):
            self.send_emails(len(self.pushes), stats, error_line)

        output = self.option("output")
        if output:
            # Build stats for CSV
            with open(output, "w") as csvfile:
                writer = csv.DictWriter(
                    csvfile,
                    fieldnames=[
                        "revision",
                        "date",
                        "classification",
                        "backedout",
                        "error_type",
                        "error_message",
                    ],
                )
                writer.writeheader()
                writer.writerows([self.build_stats(push) for push in self.pushes])
            self.line(
                f"<info>Written stats for {len(self.pushes)} pushes in {output}</info>"
            )

    def build_stats(self, push):
        """
        Build a dict with statistics relevant for a push
        """
        classification = self.classifications.get(push)
        error = self.errors.get(push)

        return {
            "revision": push.rev,
            "date": push.date,
            "classification": classification or "error",
            "backedout": push.backedout if classification else "",
            "error_type": error.__class__.__name__ if error else "",
            "error_message": str(error) if error else "",
        }

    def log_pushes(self, status, backedout):
        """
        Display stats for all pushes in a given classification state + backout combination
        """
        assert isinstance(status, PushStatus)
        assert isinstance(backedout, bool)

        nb = len(
            [
                push
                for push in self.pushes
                if self.classifications.get(push) == status
                and push.backedout == backedout
            ]
        )
        verb = "were" if backedout else "weren't"
        line = f"{nb} out of {len(self.pushes)} pushes {verb} backed-out by a sheriff and were classified as {status.name}."
        self.line(line)

        return line

    def send_emails(self, total, stats, error_line):
        today = datetime.datetime.strftime(datetime.datetime.now(), "%Y-%m-%d")

        stats = "\n".join([f"- {stat}" for stat in stats])

        environment = self.option("environment")
        notify_email(
            emails=config.get("emails", {}).get("monitoring"),
            subject=f"{environment} classify-eval report generated the {today}",
            content=EMAIL_CLASSIFY_EVAL.format(
                today=today,
                total=total,
                error_line=f"**{error_line}**" if error_line else "",
                stats=stats,
            ),
        )


class ClassifyPerfCommand(Command):
    name = "push perf"
    description = (
        "Generate a CSV file with performance stats for all classification tasks"
    )
    arguments = [
        argument(
            "branch",
            description="Branch the push belongs to (e.g autoland, try, etc).",
            optional=True,
            default="autoland",
        )
    ]
    options = [
        option(
            "environment",
            description="Environment in which the analysis is running (testing, production, ...).",
            flag=False,
            default="testing",
        ),
        option(
            "output",
            description="Output CSV file path.",
            flag=False,
            default="perfs.csv",
        ),
    ]

    REGEX_ROUTE = re.compile(
        r"^index.project.mozci.classification.([\w\-]+).(revision|push).(\w+)$"
    )

    def handle(self) -> None:
        environment = self.option("environment")
        output = self.option("output")

        # Aggregate stats for completed tasks processed by the hook
        stats = [
            self.parse_task_status(task_status)
            for group_id in self.list_groups_from_hook(
                "project-mozci", f"decision-task-{environment}"
            )
            for task_status in self.list_classification_tasks(group_id)
        ]

        # Dump stats as CSV file
        with open(output, "w") as csvfile:
            writer = csv.DictWriter(
                csvfile,
                fieldnames=[
                    "branch",
                    "push",
                    "revision",
                    "task_id",
                    "created",
                    "time_taken",
                ],
            )
            writer.writeheader()
            writer.writerows(stats)

        self.line(f"<info>Written stats for {len(stats)} tasks in {output}</info>")

    def parse_routes(self, routes):
        """Find revision from task routes"""

        def _match(route):
            res = self.REGEX_ROUTE.search(route)
            if res:
                return res.groups()

        # Extract branch+name+value from the routes
        # and get 3 separated lists to check those values
        branches, keys, values = zip(*filter(None, map(_match, routes)))

        # We should only have one branch
        branches = set(branches)
        assert len(branches) == 1, f"Multiple branches detected: {branches}"

        # Output single branch, revision and push id
        data = dict(zip(keys, values))
        assert "revision" in data, "Missing revision route"
        assert "push" in data, "Missing push route"
        return branches.pop(), data["revision"], int(data["push"])

    def parse_task_status(self, task_status):
        """Extract identification and time spent for each classification task"""

        def date(x):
            return datetime.datetime.strptime(x, "%Y-%m-%dT%H:%M:%S.%fZ")

        out = {
            "task_id": task_status["status"]["taskId"],
            "created": task_status["task"]["created"],
            "time_taken": sum(
                (date(run["resolved"]) - date(run["started"])).total_seconds()
                for run in task_status["status"]["runs"]
                if run["state"] == "completed"
            ),
        }
        out["branch"], out["revision"], out["push"] = self.parse_routes(
            task_status["task"]["routes"]
        )
        return out

    def list_groups_from_hook(self, group_id, hook_id):
        """List all decision tasks from the specified hook"""
        hooks = taskcluster.Hooks(get_taskcluster_options())
        fires = hooks.listLastFires(group_id, hook_id).get("lastFires", [])

        # Setup CLI progress bar
        progress = self.progress_bar(len(fires))
        progress.set_format("verbose")

        # Provide the decision task ID as it's the same value for group ID
        for fire in fires:
            yield fire["taskId"]

            progress.advance()

        # Cleanup progress bar
        progress.finish()

    def list_classification_tasks(self, group_id):
        # Check cache first
        cache_key = f"perf/task_group/{group_id}"
        tasks = config.cache.get(cache_key, [])

        if not tasks:
            queue = taskcluster.Queue(get_taskcluster_options())
            token = False
            try:
                # Support pagination using continuation token
                while token is not None:
                    query = {"continuationToken": token} if token else {}
                    results = queue.listTaskGroup(group_id, query=query)
                    tasks += results.get("tasks")
                    token = results.get("continuationToken")
            except TaskclusterRestFailure as e:
                # Skip expired task groups
                if e.status_code == 404:
                    return

                raise
        else:
            logger.debug("From cache", cache_key)

        for task_status in tasks:
            task_id = task_status["status"]["taskId"]

            # Skip decision task
            if task_id == group_id:
                continue

            # Only provide completed tasks
            if task_status["status"]["state"] != "completed":
                logger.debug(f"Skip not completed task {task_id}")
                continue

            yield task_status

        # Cache all tasks if all completed
        if all(t["status"]["state"] == "completed" for t in tasks):
            config.cache.add(cache_key, tasks, int(config["cache"]["retention"]))



# -*- coding: utf-8 -*-
import concurrent.futures
import copy
import itertools
import json
import math
from collections import Counter, defaultdict
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from typing import Dict, Iterator, List, Optional, Set, Tuple, Union

from loguru import logger

from mozci import config, data
from mozci.data.sources.treeherder import JobUnavailable, TreeherderClientSource
from mozci.errors import (
    ChildPushNotFound,
    MissingDataError,
    ParentPushNotFound,
    PushNotFound,
    SourcesNotFound,
)
from mozci.task import (
    GroupSummary,
    LabelSummary,
    RunnableSummary,
    Status,
    Task,
    TestTask,
    get_configuration,
)
from mozci.util.defs import FAILURE_CLASSES, TASK_FINAL_STATES
from mozci.util.hgmo import HgRev, parse_bugs
from mozci.util.memoize import memoize, memoized_property
from mozci.util.taskcluster import get_task

BASE_INDEX = "gecko.v2.{branch}.revision.{rev}"

# The maximum number of parents or children to look for previous/next task runs,
# when the task did not run on the currently considered push.
MAX_DEPTH = config.get("maxdepth", 20)


class PushStatus(Enum):
    GOOD = 0
    BAD = 1
    UNKNOWN = 2


@dataclass
class Regressions:
    # These 3 attributes are dicts of list of tasks
    # each item being a single group, with its failing tasks
    real: Dict[str, List[TestTask]]
    intermittent: Dict[str, List[TestTask]]
    unknown: Dict[str, List[TestTask]]


@dataclass
class ToRetriggerOrBackfill:
    # These 3 attributes are dicts of list of tasks
    # each item being a single group, with its failing tasks
    real_retrigger: Dict[str, List[TestTask]]
    intermittent_retrigger: Dict[str, List[TestTask]]
    backfill: Dict[str, List[TestTask]]


def build_group_summaries(tasks) -> Dict[str, GroupSummary]:
    groups = defaultdict(list)

    for task in tasks:
        if not isinstance(task, TestTask):
            continue

        for group in task.groups:
            groups[group].append(task)

    return {group: GroupSummary(group, tasks) for group, tasks in groups.items()}


class Push:
    """A representation of a single push.

    Args:
        revs (list): List of revisions of commits in the push (top-most is the first element).
        branch (str): Branch to look on (default: autoland).
    """

    # static thread pool, to avoid spawning threads too often.
    THREAD_POOL_EXECUTOR = concurrent.futures.ThreadPoolExecutor()

    def __init__(self, revs, branch="autoland"):
        if isinstance(revs, str):
            # Direct usage of a single revision reference
            self._revs = None
            head_revision = revs
            self._bugs = None
        elif isinstance(revs, list) and len(revs) > 0:
            if all(map(lambda r: isinstance(r, dict), revs)):
                # We should get detailed revision objects here
                # and get the list of Bugzilla Bug Ids from the description
                self._bugs = set(
                    itertools.chain(*[parse_bugs(rev.get("desc", "")) for rev in revs])
                )

                self._revs = [r["node"] for r in revs]
            else:
                # Support list of changeset IDs
                self._bugs = None
                self._revs = revs

            head_revision = self._revs[0]
        else:
            raise NotImplementedError(f"Cannot process revisions: {revs}")

        self.branch = branch
        self._hgmo = HgRev.create(head_revision, branch=self.branch)

        self._id = None
        self._date = None
        self._author = None

        # Need to use full hash in the index.
        if len(head_revision) == 40:
            self.rev = head_revision
        else:
            self.rev = self._hgmo.node

        self.index = BASE_INDEX.format(branch=self.branch, rev=self.rev)
        # Unique identifier for a Push across branches
        self.push_uuid = "{}/{}".format(self.branch, self.rev)

    @property
    def revs(self):
        if not self._revs:
            self._revs = [c["node"] for c in self._hgmo.changesets]
        return self._revs

    @memoized_property
    def backedoutby(self):
        """The revision of the commit which backs out this one or None.

        Returns:
            str or None: The commit revision which backs this push out (or None).
        """
        return self._hgmo.backedoutby

    @property
    def backedout(self):
        """Whether the push was backed out or not.

        Returns:
            bool: True if this push was backed out.
        """
        return bool(self.backedoutby)

    @property
    def bugs(self):
        """The bugs associated with the commits of this push.

        Returns:
            set: A set of bug IDs.
        """
        if self._bugs:
            return self._bugs

        self._bugs = self._hgmo.bugs
        return self._bugs

    @property
    def date(self):
        """The push date.

        Returns:
            int: The push date in ms since the epoch.
        """
        if self._date:
            return self._date

        self._date = self._hgmo.pushdate
        return self._date

    @property
    def id(self):
        """The push id.

        Returns:
            int: The push id.
        """
        if self._id:
            return self._id

        self._id = self._hgmo.pushid
        return self._id

    @property
    def author(self):
        """The push author.

        Returns:
            str: The push author.
        """
        if self._author:
            return self._author

        self._author = self._hgmo.pushauthor
        return self._author

    @property
    def is_finalized(self):
        """The push is finished or not.

        Returns:
            bool: True if the push is considered finalized (> 1 day old), else False.
        """
        yesterday = (datetime.now() - timedelta(days=1)).timestamp()
        return self.date < yesterday

    def create_push(self, push_id):
        result = HgRev.load_json_push(self.branch, push_id)

        push = Push(result["changesets"][::-1], branch=self.branch)
        # avoids the need to query hgmo to find this info
        push._id = push_id
        push._date = result["date"]

        return push

    @memoized_property
    def parent(self):
        """Returns the parent push of this push.

        Returns:
            Push: A `Push` instance representing the parent push.

        Raises:
            :class:`~mozci.errors.ParentPushNotFound`: When no suitable parent
                push can be detected.
        """
        # Mozilla-unified and try allow multiple heads, so we can't rely on
        # `self.id - 1` to be the parent.
        if self.branch not in ("mozilla-unified", "try"):
            return self.create_push(self.id - 1)

        changesets = [c for c in self._hgmo.changesets if c.get("phase") == "draft"]
        if not changesets:
            # Supports mozilla-unified as well as older automationrelevance
            # files that don't contain the phase.
            changesets = self._hgmo.changesets

        parents = changesets[0]["parents"]
        if len(parents) > 1:
            raise ParentPushNotFound(
                "merge commits are unsupported", rev=self.rev, branch=self.branch
            )

        # Search for this revision in the following repositories. We search
        # autoland last as it would run the fewest tasks, so a push from one of
        # the other repositories would be preferable.
        branches = ("mozilla-central", "mozilla-beta", "mozilla-release", "autoland")
        found_on = []
        parent_rev = parents[0]
        for branch in branches:
            try:
                hgmo = HgRev.create(parent_rev, branch=branch)
                head = hgmo.pushhead
            except PushNotFound:
                continue

            found_on.append(branch)

            # Revision exists in repo but is not the 'pushhead', so keep searching.
            if head != parent_rev:
                continue

            return Push(parent_rev, branch=branch)

        if found_on:
            branches = found_on
            reason = "was not a push head"
        else:
            # This should be extremely rare (if not impossible).
            reason = "was not found"

        branches = [
            b[len("mozilla-") :] if b.startswith("mozilla-") else b for b in branches
        ]
        msg = f"parent revision '{parent_rev[:12]}' {reason} on any of {', '.join(branches)}"
        raise ParentPushNotFound(msg, rev=self.rev, branch=self.branch)

    @memoized_property
    def child(self):
        """Returns the child push of this push.

        Returns:
            Push: A `Push` instance representing the child push.

        Raises:
            :class:`~mozci.errors.ChildPushNotFound`: When no suitable child
                push can be detected.
        """
        if self.branch not in ("mozilla-unified", "try"):
            try:
                return self.create_push(self.id + 1)
            except PushNotFound as e:
                raise ChildPushNotFound(
                    "child push does not exist", rev=self.rev, branch=self.branch
                ) from e

        raise ChildPushNotFound(
            f"finding child pushes not supported on {self.branch}",
            rev=self.rev,
            branch=self.branch,
        )

    @memoized_property
    def decision_task(self):
        """A representation of the decision task.

        Returns:
            Task: A `Task` instance representing the decision task.
        """
        index = self.index + ".taskgraph.decision"
        return Task.create(index=index)

    @memoized_property
    def tasks(self):
        """All tasks that ran on the push, including retriggers and backfills.

        Returns:
            list: A list of `Task` objects.
        """
        cached_tasks = config.cache.get(f"{self.push_uuid}/tasks")
        # Push is supposedly finalized, if a cache exists we can return it as tasks should've finished running
        if self.is_finalized and cached_tasks is not None:
            return cached_tasks

        logger.debug(f"Retrieving all tasks and groups which run on {self.rev}...")

        logger.debug(f"Gathering data about tasks that run on {self.rev}...")

        # Gather data about tasks that ran on a given push.
        try:
            # TODO: Skip tasks with `retry` as result
            tasks = data.handler.get("push_tasks", branch=self.branch, rev=self.rev)
        except MissingDataError:
            return []

        # Update gathered tasks with data retrieved from the cache
        completed_cached_tasks = {}
        if cached_tasks:
            completed_cached_tasks = {
                t.id: vars(t) for t in cached_tasks if t.state in TASK_FINAL_STATES
            }
            tasks = [{**t, **completed_cached_tasks.get(t["id"], {})} for t in tasks]

        logger.debug(
            f"{len(completed_cached_tasks)} tasks were retrieved from the cache..."
        )

        logger.debug(f"Gathering task classifications for {self.rev}...")

        # Gather task classifications.
        try:
            classifications = data.handler.get(
                "push_tasks_classifications", branch=self.branch, rev=self.rev
            )
            for task in tasks:
                if task["id"] in classifications:
                    task.update(classifications[task["id"]])
        except MissingDataError:
            pass

        tasks = [Task.create(**task) for task in tasks]

        # Gather group data.
        logger.debug(f"Gathering test groups for {self.rev}...")
        done, _ = concurrent.futures.wait(
            [
                Push.THREAD_POOL_EXECUTOR.submit(
                    lambda task: task.retrieve_results(self), task
                )
                for task in tasks
                # No need to gather group data for a completed task that was already cached
                if task.id not in completed_cached_tasks and isinstance(task, TestTask)
            ],
            return_when=concurrent.futures.FIRST_EXCEPTION,
        )
        failed = set()
        for f in done:
            try:
                f.result()
            except SourcesNotFound as e:
                task = e.context["task"]
                failed.add(f"{task.id} - {task.label}")

        if failed:
            failed_str = "  \n".join(sorted(failed))
            logger.warning(
                f"Failed to get test groups from the following tasks:\n  {failed_str}"
            )

        logger.debug(f"Retrieved all tasks and groups which run on {self.rev}.")

        # Skip tier tasks greater than the tier passed in config
        tasks = [task for task in tasks if not task.tier or task.tier <= config.tier]

        # cachy's put() overwrites the value in the cache; add() would only add if its empty
        config.cache.put(
            f"{self.push_uuid}/tasks",
            tasks,
            config["cache"]["retention"],
        )

        logger.debug(f"Cached all tasks and groups which run on {self.rev}.")

        return tasks

    @property
    def task_labels(self):
        """The set of task labels that ran on this push.

        Returns:
            set: A set of task labels (str).
        """
        return set(t.label for t in self.tasks)

    @memoized_property
    def is_manifest_level(self):
        """Whether a non-default manifest loader was used for this push.

        Returns:
            bool: True if a non-default manifest loader was used.
        """
        return (
            self.decision_task.get_artifact("public/parameters.yml")[
                "test_manifest_loader"
            ]
            != "default"
        )

    @memoized_property
    def target_task_labels(self):
        """The set of all task labels that could possibly run on this push.

        Returns:
            set: A set of task labels.
        """
        return set(self.decision_task.get_artifact("public/target-tasks.json"))

    @memoized_property
    def scheduled_task_labels(self):
        """The set of task labels that were originally scheduled to run on this push.

        This excludes backfills and Add New Jobs.

        Returns:
            set: A set of task labels (str).
        """
        tasks = self.decision_task.get_artifact("public/task-graph.json").values()
        return {t["label"] for t in tasks}

    @property
    def unscheduled_task_labels(self):
        """The set of task labels from tasks that were not originally scheduled on
        the push (i.e they were scheduled via backfill or Add New Jobs).

        Returns:
            set: A set of task labels (str).
        """
        return self.task_labels - self.scheduled_task_labels

    @memoized_property
    def config_group_summaries(self):
        """All group summaries, on given configurations, combining retriggers.

        Returns:
            dict: A dictionary of the form {(<configuration>, <group>): [<GroupSummary>]}.
        """
        config_groups = defaultdict(list)

        for task in self.tasks:
            if not isinstance(task, TestTask):
                continue

            for group in task.groups:
                config_groups[(task.configuration, group)].append(task)

        return {
            config_group: GroupSummary(config_group[1], tasks)
            for config_group, tasks in config_groups.items()
        }

    @memoized_property
    def group_summaries(self):
        """All group summaries combining retriggers.

        Returns:
            dict: A dictionary of the form {<group>: [<GroupSummary>]}.
        """
        return build_group_summaries(self.tasks)

    @memoized_property
    def label_summaries(self) -> Dict[str, LabelSummary]:
        """All label summaries combining retriggers.

        Returns:
            dict: A dictionary of the form {<label>: [<LabelSummary>]}.
        """
        labels = defaultdict(list)
        for task in self.tasks:
            # We can't consider tasks that were chunked in the taskgraph for finding label-level regressions
            # because tasks with the same name on different pushes might contain totally different tests.
            if task.is_tests_grouped:
                continue

            labels[task.label].append(task)
        return {label: LabelSummary(label, tasks) for label, tasks in labels.items()}

    @memoized_property
    def duration(self):
        """The total duration of all tasks that ran on the push.

        Returns:
            int: Runtime in hours.
        """
        return int(sum(t.duration for t in self.tasks) / 3600)

    @memoized_property
    def scheduled_duration(self):
        """The total runtime of tasks excluding retriggers and backfills.

        Returns:
            int: Runtime in hours.
        """
        seen = set()
        duration = 0
        for task in self.tasks:
            if task.label not in self.scheduled_task_labels:
                continue

            if task.label in seen:
                continue

            seen.add(task.label)
            duration += task.duration

        return int(duration / 3600)

    def is_group_running(self, group):
        """Checks if the provided group is still running on this push.

        Returns:
            bool: True, if the group is still running.
                  False, if the group is completed.
        """
        running_tasks = [
            task for task in self.tasks if task.state not in TASK_FINAL_STATES
        ]

        group_types = set()
        for task in group.tasks:
            suite = task.suite
            assert suite is not None, (
                f"Couldn't parse suite for {task.label} ({task.id})"
            )
            group_types.add(suite)

        if all(task.is_tests_grouped for task in group.tasks):
            for task in running_tasks:
                if task.suite not in group_types:
                    continue

                task_def = get_task(task.id)
                test_paths = json.loads(
                    task_def["payload"]
                    .get("env", {})
                    .get("MOZHARNESS_TEST_PATHS", "{}")
                )
                if group.name in {
                    name for names in test_paths.values() for name in names
                }:
                    return True
            return False

        running_types = {task.suite for task in running_tasks}
        return not group_types.isdisjoint(running_types)

    def _is_classified_as_cause(self, first_appearance_push, classifications):
        """Checks a 'fixed by commit' classification to figure out what push it references.

        Returns:
            bool or None: True, if the classification references this push.
                          False, if the classification references another push.
                          None, if it is not clear what the classification references.
        """
        fixed_by_commit_classification_notes = set(
            n[:12]
            for c, n in classifications
            if c == "fixed by commit"
            if n is not None
        )

        if len(fixed_by_commit_classification_notes) == 0:
            return None

        # If the failure was classified as fixed by commit, and the fixing commit
        # is a backout of the current push, it is definitely a regression of the
        # current push.
        # If the failure was classified as fixed by commit, and the fixing commit
        # is a backout of another push, it is definitely not a regression of the
        # current push.
        # Unless some condition holds which makes us doubt about the correctness of the
        # classification.
        # - the backout commit also backs out one of the commits of this push;
        # - the other push backed-out by the commit which is mentioned in the classification
        #   is landed after the push where the failure occurs (so, it can't have caused it);
        # - the backout push also contains a commit backing out one of the commits of this push.
        for classification_note in fixed_by_commit_classification_notes:
            fix_hgmo = HgRev.create(classification_note, branch=self.branch)
            try:
                fix_hgmo.backouts
            except PushNotFound:
                logger.warning(
                    f"Classification note ({classification_note}) references a revision which does not exist on push {first_appearance_push.rev}"
                )
                return None

            self_fix = None
            other_fixes = set()

            # If the backout commit also backs out one of the commits of this push, then
            # we can consider it as a regression of this push.
            # NOTE: this should never happen in practice because of current development
            # practices.
            for backout, backedouts in fix_hgmo.backouts.items():
                if backout[:12] != classification_note[:12]:
                    continue

                if any(
                    rev[:12] in {backedout[:12] for backedout in backedouts}
                    for rev in self.revs
                ):
                    self_fix = backout[:12]
                    break

            # Otherwise, if the backout push also contains the backout commit of this push,
            # we can consider it as a regression of this push.
            if self.backedout:
                for backout in fix_hgmo.backouts:
                    if backout[:12] == self.backedoutby[:12]:
                        self_fix = backout[:12]
                        break

            # If one of the commits in the backout push is a bustage fix, then we could
            # consider it as a regression of this push.
            if self_fix is None:
                for bug in self.bugs:
                    if bug in fix_hgmo.bugs_without_backouts:
                        self_fix = fix_hgmo.bugs_without_backouts[bug][:12]
                        break

            # Otherwise, as long as the commit which was backed-out was landed **before**
            # the appearance of this failure, we can be sure it was its cause and so
            # the current push is not at fault.
            # TODO: We should actually check if the failure was already happening in the parents
            # and compare the backout push ID with the the ID of first parent where it failed.
            for backout, backedouts in fix_hgmo.backouts.items():
                if self.backedout and backout[:12] == self.backedoutby[:12]:
                    continue

                if any(
                    HgRev.create(backedout, branch=self.branch).pushid
                    <= first_appearance_push.id
                    for backedout in backedouts
                ):
                    other_fixes.add(backout[:12])

            # If the backout push contains a bustage fix of another push, then we could
            # consider it as a regression of another push.
            if len(fix_hgmo.bugs_without_backouts) > 0:
                other_parent = first_appearance_push
                for other_parent in other_parent._iterate_parents(MAX_DEPTH):
                    if other_parent != self:
                        for bug in other_parent.bugs:
                            if bug in fix_hgmo.bugs_without_backouts:
                                other_fixes.add(
                                    fix_hgmo.bugs_without_backouts[bug][:12]
                                )

            if self_fix and other_fixes:
                # If the classification points to a commit in the middle of the backout push and not the backout push head,
                # we can consider the classification to be correct.
                if (
                    self_fix != fix_hgmo.pushhead[:12]
                    and classification_note[:12] == self_fix
                    and classification_note[:12] not in other_fixes
                ):
                    return True
                elif any(
                    other_fix != fix_hgmo.pushhead[:12]
                    and classification_note[:12] == other_fix
                    and classification_note[:12] != self_fix
                    for other_fix in other_fixes
                ):
                    return False

                return None

            if self_fix:
                return True

            if other_fixes:
                return False

        return None

    def _iterate_children(self, max_depth=None):
        other = self
        for i in itertools.count():
            yield other

            # Optimization to load child json-pushes data in a single query (up
            # to MAX_DEPTH at a time).
            next_id = other.id + 1
            if next_id not in HgRev.JSON_PUSHES_CACHE:
                depth = max_depth or MAX_DEPTH
                HgRev.load_json_pushes_between_ids(
                    self.branch, other.id, next_id + depth - i
                )

            try:
                other = other.child
            except ChildPushNotFound:
                break

            if max_depth is not None and i == max_depth:
                break

    def _iterate_parents(self, max_depth=None):
        other = self
        for i in itertools.count():
            yield other

            # Optimization to load parent json-pushes data in a single query (up
            # to max_depth at a time).
            prev_id = other.id - 1
            if prev_id not in HgRev.JSON_PUSHES_CACHE:
                depth = max_depth or MAX_DEPTH
                HgRev.load_json_pushes_between_ids(
                    self.branch, max(prev_id - 1 - depth + i, 0), prev_id
                )

            try:
                other = other.parent
            except ParentPushNotFound:
                break

            if max_depth is not None and i == max_depth:
                break

    def _iterate_failures(
        self, runnable_type: str, max_depth: Optional[int] = None
    ) -> Iterator[
        Tuple[
            "Push",
            Dict[str, Tuple[float, RunnableSummary]],
            Dict[str, Tuple[float, RunnableSummary]],
            Dict[str, List[Optional[bool]]],
        ]
    ]:
        ever_passing_runnables = set()
        passing_runnables = set()
        candidate_regressions = {}

        first_appearance = {}

        classified_as_cause: Dict[str, List[Optional[bool]]] = defaultdict(list)

        count = 0
        for other in self._iterate_children(max_depth):
            for name, summary in getattr(other, f"{runnable_type}_summaries").items():
                # test-verify is special, we don't want to look at children pushes.
                if (
                    self != other
                    and runnable_type == "label"
                    and ("test-verify" in name or "test-coverage" in name)
                ):
                    break

                if summary.status == Status.PASS:
                    ever_passing_runnables.add(name)
                    if name not in candidate_regressions:
                        passing_runnables.add(name)
                    continue

                if all(c not in FAILURE_CLASSES for c, n in summary.classifications):
                    classified_as_cause[name].append(None)
                    if name not in candidate_regressions:
                        passing_runnables.add(name)
                    continue

                if name not in first_appearance:
                    first_appearance[name] = other

                is_classified_as_cause = self._is_classified_as_cause(
                    first_appearance[name], summary.classifications
                )
                if is_classified_as_cause is True:
                    classified_as_cause[name].append(True)
                elif is_classified_as_cause is False:
                    classified_as_cause[name].append(False)
                    continue

                if name in candidate_regressions:
                    # It failed in one of the pushes between the current and its
                    # children, we don't want to increase the previous distance.
                    continue

                candidate_regressions[name] = (float(count), summary)

            adjusted_candidate_regressions = copy.deepcopy(candidate_regressions)

            for name in candidate_regressions.keys():
                # If the classifications are all either 'intermittent' or 'fixed by commit' pointing
                # to this push, and the last classification is 'fixed by commit', then consider it a
                # sure regression. We are assuming sheriff's information might be wrong at first and
                # adjusted later.
                if (
                    len(classified_as_cause[name]) > 0
                    and all(
                        result is True or result is None
                        for result in classified_as_cause[name]
                    )
                    and classified_as_cause[name][-1] is True
                ):
                    adjusted_candidate_regressions[name] = (
                        -math.inf,
                        candidate_regressions[name][1],
                    )

                # If there is at least one classification that points to another push and no classification
                # that points to this push, remove the regerssion from the candidate list.
                elif any(
                    result is False for result in classified_as_cause[name]
                ) and not any(result is True for result in classified_as_cause[name]):
                    del adjusted_candidate_regressions[name]

                # If the runnable passed first and failed in a child push and there is no classification pointing
                # to this push or there is at least one pointing to another push, remove the regression from the
                # candidate list.
                elif name in passing_runnables and (
                    not any(result is True for result in classified_as_cause[name])
                    or any(result is False for result in classified_as_cause[name])
                ):
                    del adjusted_candidate_regressions[name]

                # If the runnable passed in any child push and the classification for the last seen failure is
                # 'intermittent', remove the regression from the candidate list.
                elif (
                    len(classified_as_cause[name]) > 0
                    and name in ever_passing_runnables
                    and classified_as_cause[name][-1] is None
                ):
                    del adjusted_candidate_regressions[name]

                # If all classifications are 'intermittent', remove the regression from the candidate list.
                elif len(classified_as_cause[name]) > 0 and all(
                    result is None for result in classified_as_cause[name]
                ):
                    del adjusted_candidate_regressions[name]

            yield (
                other,
                adjusted_candidate_regressions,
                candidate_regressions,
                classified_as_cause,
            )
            count += 1

    def get_candidate_regressions(
        self, runnable_type: str
    ) -> Dict[str, Tuple[float, RunnableSummary]]:
        """Retrieve the set of "runnables" that are regression candidates for this push.

        A "runnable" can be any group of tests, e.g. a label, a manifest across platforms,
        a manifest on a given platform.

        A candidate regression is any runnable for which at least one
        associated task failed (therefore including intermittents), and which
        is either not classified or fixed by commit.

        Returns:
            set: Set of runnable names (str).
        """
        logger.debug(f"Retrieving candidate regressions for {self.rev}...")

        max_depth = None if self.backedout or self.bustage_fixed_by else MAX_DEPTH

        for other, candidate_regressions, _, _ in self._iterate_failures(
            runnable_type, max_depth
        ):
            # Break early if we reached the backout of this push, since any failure
            # after that won't be blamed on this push.
            try:
                next_child = other.child
            except ChildPushNotFound:
                break

            if self.branch != "try" and (
                self.backedoutby in next_child.revs
                or self.bustage_fixed_by in next_child.revs
            ):
                logger.debug(
                    f"Reached a backout/bustage fix of {self.rev}, stop looking for failures in children."
                )

                # Runnables that still fail after the backout, can't be considered
                # regressions of the push.
                # NOTE: There could be another push in between the push of interest and its
                # backout that causes another failure in the same runnable, but it is very
                # unlikely (especially for finer granularities, such as "group").
                for name, summary in getattr(
                    next_child, f"{runnable_type}_summaries"
                ).items():
                    if (
                        name in candidate_regressions
                        and summary.status != Status.PASS
                        and all(
                            c in FAILURE_CLASSES for c, n in summary.classifications
                        )
                    ):
                        del candidate_regressions[name]

                break

        return candidate_regressions

    @memoized_property
    def bustage_fixed_by(self) -> Optional[str]:
        """The revision of the commit which 'bustage fixes' this one or None.

        We detect if a push was 'bustage fixed' with a simple heuristic:

            - there is a close enough child push where the task/group passes;
            - the child push where the task/group passes is associated to the same bug
              as the push of interest.

        Returns:
            str or None: The commit revision which 'bustage fixes' this push (or None).
        """
        if self.backedout:
            return None

        def fix_same_bugs(push1, push2):
            return len(push1.bugs & push2.bugs) > 0

        # Skip checking regressions if we can't find any possible candidate.
        possible_bustage_fixes = set(
            other
            for other in self._iterate_children(MAX_DEPTH)
            if self != other and fix_same_bugs(self, other)
        )
        if len(possible_bustage_fixes) == 0:
            return None

        def find(runnable_type: str) -> Optional[str]:
            for other, candidate_regressions, _, _ in self._iterate_failures(
                runnable_type, MAX_DEPTH
            ):
                if other == self or other not in possible_bustage_fixes:
                    continue

                other_summaries = getattr(other, f"{runnable_type}_summaries")

                if any(
                    name in other_summaries
                    and other_summaries[name].status == Status.PASS
                    for name in candidate_regressions
                ):
                    return other.rev

                possible_bustage_fixes.remove(other)
                if len(possible_bustage_fixes) == 0:
                    break

            return None

        return find("label") or find("config_group")

    @memoize
    def get_regressions(
        self,
        runnable_type: str,
        historical_analysis: bool = True,
        assume_unclassified_is_intermittent: bool = False,
    ) -> Dict[str, int]:
        """All regressions, both likely and definite.

        Each regression is associated with an integer, which is the number of
        parent and children pushes that didn't run the runnable. A count of 0 means
        the runnable failed on the current push and passed on the previous push.
        A count of 3 means there were three pushes between the failure and the
        last time the task passed (so any one of them could have caused it).
        A count of MAX_DEPTH means that the maximum number of parents were
        searched without finding the task and we gave up.
        If the parameter `assume_unclassified_is_intermittent` is set to True, previous
        failures that have not been classified will be considered as intermittent.

        Returns:
            dict: A dict of the form {<str>: <int>}.
        """
        regressions: Dict[str, int] = {}

        # If the push was not backed-out and was not "bustage fixed", it can't
        # have caused regressions.
        if (
            historical_analysis
            and self.branch != "try"
            and not self.backedout
            and not self.bustage_fixed_by
        ):
            return regressions

        for name, (count, failure_summary) in self.get_candidate_regressions(
            runnable_type
        ).items():
            other = self.parent
            prior_regression = False

            # test-verify is special, we can assume the test-verify task is not the same as the one
            # in the parent pushes.
            if (
                runnable_type != "label"
                or "test-verify" not in name
                or "test-coverage" not in name
            ):
                found_in_parent = False
                for other in other._iterate_parents(MAX_DEPTH):
                    if count < 0:
                        break

                    runnable_summaries = getattr(other, f"{runnable_type}_summaries")

                    if name in runnable_summaries:
                        found_in_parent = True
                        summary = runnable_summaries[name]

                        # If the failure is not intermittent...
                        if not failure_summary.is_intermittent:
                            # ...and it failed permanently in the first parent where it ran, it is a
                            # prior regression.
                            # Otherwise, if it passed or was intermittent, it is likely not a prior
                            # regression.
                            if (
                                summary.status != Status.PASS
                                and not summary.is_intermittent
                                and (
                                    not assume_unclassified_is_intermittent
                                    or all(
                                        c != "not classified"
                                        for c, n in summary.classifications
                                    )
                                )
                            ):
                                prior_regression = True

                            break

                        # If the failure is intermittent and it failed intermittently in a close
                        # parent too, it is likely a prior regression.
                        # We need to explore the parent's parents instead if it passed or failed
                        # consistently in the parent.
                        elif summary.is_intermittent:
                            prior_regression = True
                            break

                    if not found_in_parent:
                        count += 1

            # Given that our "bustage fix" detection is a heuristic which might fail, we
            # penalize regressions for pushes which weren't backed-out by doubling their count
            # (basically, we consider the push to be further away from the failure, which makes
            # it more likely to fall outside of MAX_DEPTH).
            if historical_analysis and self.branch != "try" and not self.backedout:
                count *= 2

            # Also penalize cases where the status was intermittent.
            if failure_summary.is_intermittent:
                count *= 2

            # if confirmed = True; then assume regression
            # if confirmed = False; then assume it's very unlikely a regression
            if isinstance(failure_summary, GroupSummary):
                if failure_summary.is_confirmed_failure():
                    count = 0
                elif failure_summary.is_confirmed_failure() is False:
                    count = MAX_DEPTH + 1

            if not prior_regression and count <= MAX_DEPTH:
                regressions[name] = int(count) if count > 0 else 0

        return regressions

    def get_possible_regressions(
        self, runnable_type: str, historical_analysis: bool = True
    ) -> Set[str]:
        """The set of all runnables that may have been regressed by this push.

        A possible regression is a candidate_regression that didn't run on one or
        more parent pushes.

        Returns:
            set: Set of runnables (str).
        """
        return set(
            name
            for name, count in self.get_regressions(
                runnable_type, historical_analysis
            ).items()
            if count > 0
        )

    def get_likely_regressions(
        self, runnable_type: str, historical_analysis: bool = True
    ) -> Set[str]:
        """The set of all runnables that were likely regressed by this push.

        A likely regression is a candidate_regression that both ran and passed
        on the immediate parent push. It still isn't a sure thing as the task
        could be intermittent.

        Returns:
            set: Set of runnables (str).
        """
        return set(
            name
            for name, count in self.get_regressions(
                runnable_type, historical_analysis
            ).items()
            if count == 0
        )

    def should_retrigger_build(self, task: Task) -> bool:
        """Extra logic to determine if a build task of the current push should be retriggered."""

        if task.tier not in (1, 2):
            return False

        # There should be only one task on the current push with the same label, as we only want to retrigger once
        if sum(1 for t in self.tasks if t.label == task.label) > 1:
            return False

        return True

    def check_build_regressions(self) -> list[Task]:
        logger.info(f"Fetched {len(self.tasks)} tasks for push {self.id}.")

        # Try to identify a potential regressions from the failed build
        potential_regressions = self.get_regressions(
            "label",
            historical_analysis=False,
            assume_unclassified_is_intermittent=True,
        )

        # Map labels to tasks again
        build_regressions = [
            (task, past_occurrences)
            for label, past_occurrences in potential_regressions.items()
            if (task := next((t for t in self.tasks if t.label == label), None))
            and task.is_build_failure()
        ]
        if not build_regressions:
            logger.info("No build regression detected.")
            return []

        new_regressions = sum(
            past_occurrences == 0 for _, past_occurrences in build_regressions
        )
        logger.info(
            f"Detected {len(build_regressions)} build tasks that may contain a regression "
            f"({new_regressions} were never seen before)."
        )

        tasks_to_retrigger = [
            task for task, _ in build_regressions if self.should_retrigger_build(task)
        ]
        if not tasks_to_retrigger:
            logger.info("No build task should be retriggered.")
        logger.info(f"{len(tasks_to_retrigger)} build tasks should be retriggered.")
        return tasks_to_retrigger

    def classify_regressions(
        self,
        intermittent_confidence_threshold: float = 0.8,
        real_confidence_threshold: float = 0.9,
        use_possible_regressions: bool = False,
        unknown_from_regressions: bool = True,
        consider_children_pushes_configs: bool = True,
        cross_config_counts: Optional[Tuple[int, int]] = (2, 2),
        consistent_failures_counts: Optional[Tuple[int, int]] = (2, 3),
    ) -> Tuple[Regressions, ToRetriggerOrBackfill]:
        """
        Use group classification data from bugbug to classify all likely
        regressions into three categories: real, intermittent or unknown failures

        Output:
        A dict with several lists of task groups:
          - real: set of tasks that are definitely real failures
          - intermittent: set of tasks that are definitely intermittents
          - unknown: set of tasks we don't have enough information about
        And a dict with several lists of task groups:
          - real_retrigger: set of tasks to retrigger with consistent_failures_counts[1]
          - intermittent_retrigger: set of tasks to retrigger with consistent_failures_counts[0]
          - backfill: set of tasks to backfill
        """
        cache_prefix = f"{self.push_uuid}/classify_group_tasks/"

        # Retrieve test selection data for that push, from cache or bugbug
        bugbug_selection = config.cache.get(cache_prefix + "test_selection")
        if bugbug_selection is None:
            bugbug_selection = self.get_test_selection_data()

            config.cache.put(
                cache_prefix + "test_selection",
                bugbug_selection,
                config["cache"]["retention"],
            )

        # Fetch likely group regressions for that push from treeherder + Taskcluster
        # We do not cache these results as we might want to analyze in-progress
        # pushes and keep updating these values
        likely_regressions = self.get_likely_regressions("group", False)
        possible_regressions = self.get_possible_regressions("group", False)
        groups_regressions = likely_regressions
        if use_possible_regressions:
            groups_regressions |= possible_regressions

        # Get task groups with high and low confidence from bugbug scheduling
        groups_high = {
            g
            for g, confidence in bugbug_selection["groups"].items()
            if confidence >= real_confidence_threshold
        }
        groups_low = {
            g
            for g, confidence in bugbug_selection["groups"].items()
            if confidence < intermittent_confidence_threshold
        }
        logger.debug(f"Got {len(groups_high)} groups with high confidence")
        logger.debug(f"Got {len(groups_low)} groups with low confidence")

        # Classify task groups regarding cross config failure and overall failure
        # - if a group is failing in all tasks, then it is a "cross-config" failure
        # - if a group is failing only in some tasks but not all, then it is not a "cross-config" failure
        # Consider groups in following pushes too in case there is a failure due to this push in a child push.
        push_groups = self.group_summaries.values()
        if consider_children_pushes_configs:
            all_groups = build_group_summaries(
                sum((push.tasks for push in self._iterate_children(MAX_DEPTH)), [])
            ).values()
        else:
            all_groups = push_groups

        groups_confirmed = {g.name: g.is_confirmed_failure() for g in all_groups}

        groups_consistent_failure = {
            g.name
            for g in all_groups
            if groups_confirmed[g.name]
            or (
                groups_confirmed[g.name] is None
                and g.is_consistent_failure(
                    cross_config_counts[1] if cross_config_counts is not None else None,
                    consistent_failures_counts[1]
                    if consistent_failures_counts is not None
                    else None,
                )
            )
        }

        groups_non_consistent_failure = {
            g.name
            for g in all_groups
            if groups_confirmed[g.name] is False
            or (
                groups_confirmed[g.name] is None
                and g.is_consistent_failure(
                    cross_config_counts[0] if cross_config_counts is not None else None,
                    consistent_failures_counts[0]
                    if consistent_failures_counts is not None
                    else None,
                )
                is False
            )
        }

        groups_failing = {g.name for g in all_groups if g.status != Status.PASS}

        groups_no_confidence = {
            g.name
            for g in all_groups
            if g.name not in list(bugbug_selection["groups"].keys())
        }
        logger.debug(
            f"Got {len(groups_consistent_failure)} groups with cross-config or config-consistent failures"
        )
        logger.debug(
            f"Got {len(groups_non_consistent_failure)} groups with no cross-config and no config-consistent failures"
        )
        logger.debug(
            f"Got {len(groups_failing)} groups failing in the push or one of its children"
        )
        logger.debug(
            f"Got {len(groups_no_confidence)} groups without bugbug confidence"
        )

        # If failure is not classified as NEW, it's likely an intermittent.
        # We can't say the same if the failure is classified as NEW, but we can use it as a hunch.
        groups_classified_as_new = set(
            g.name
            for g in all_groups
            if g.status != Status.PASS
            if any(c == "new failure not classified" for c, n in g.classifications)
        )
        groups_not_classified_as_new = set(
            g.name
            for g in all_groups
            if g.status != Status.PASS
            if all(c != "new failure not classified" for c, n in g.classifications)
        )

        # Real failure are groups with likely regressions that were selected with high confidence
        # OR were classified as "NEW"
        # AND failing consistently across configs or in retriggers
        real_failures = (
            groups_regressions
            & groups_consistent_failure
            & groups_high.union(groups_classified_as_new)
        )

        # Intermittent failures are groups that were NOT selected (low confidence)
        # OR without any confidence from bugbug (too low confidence)
        # OR were not classified as "NEW"
        # AND are failing inconsistently across configs or in retriggers
        # Only consider groups in this push because we don't care about intermittents in other pushes.
        # We only use children pushes information to determine cross-config or config-consistent state.
        all_intermittent_failures = groups_non_consistent_failure & groups_low.union(
            groups_no_confidence
        ).union(groups_not_classified_as_new)
        intermittent_failures = (
            set(g.name for g in push_groups) & all_intermittent_failures
        )

        # Unknown failures all the remaining failing groups that are not real nor intermittent
        unknown_failures = (
            (groups_regressions if unknown_from_regressions else groups_failing)
            - real_failures
            - all_intermittent_failures
        )

        groups_still_running = set()
        for group_name in real_failures:
            for parent in self._iterate_parents(max_depth=MAX_DEPTH):
                # If the group run in a parent, then we don't care if it's still running in a grandparent.
                if group_name in parent.group_summaries:
                    break

                if parent.is_group_running(self.group_summaries[group_name]):
                    groups_still_running.add(group_name)
                    break

        logger.debug(
            f"Got {len(groups_still_running)} groups failing in this push but still running in a parent push, so we can't know if they are regressions from this push"
        )
        real_failures -= groups_still_running
        unknown_failures |= groups_still_running

        logger.debug(f"Got {len(real_failures)} real failures")
        logger.debug(f"Got {len(intermittent_failures)} intermittent failures")
        logger.debug(f"Got {len(unknown_failures)} unknown failures")

        def _map_failing_tasks(groups):
            # Link all the failing tasks on the given groups
            return {name: self.group_summaries[name].failing_tasks for name in groups}

        # See the following comment that explains how we decide the groups to retrigger/backfill
        # https://github.com/mozilla/mozci/issues/654#issuecomment-1139488070
        real_failures_to_be_retriggered = (
            (groups_regressions & groups_high.union(groups_classified_as_new))
            - groups_consistent_failure
            - groups_non_consistent_failure
        )
        real_groups_still_running = groups_still_running | {
            group_name
            for group_name in real_failures_to_be_retriggered
            if self.is_group_running(self.group_summaries[group_name])
        }
        real_failures_to_be_retriggered -= real_groups_still_running

        intermittent_failures_to_be_retriggered = (
            (
                set(g.name for g in push_groups if g.status != Status.PASS)
                & groups_low.union(groups_no_confidence).union(
                    groups_not_classified_as_new
                )
            )
            - groups_non_consistent_failure
            - groups_consistent_failure
        )
        intermittent_groups_still_running = groups_still_running | {
            group_name
            for group_name in intermittent_failures_to_be_retriggered
            if self.is_group_running(self.group_summaries[group_name])
        }
        intermittent_failures_to_be_retriggered -= intermittent_groups_still_running

        failures_to_be_backfilled = list(
            (
                possible_regressions.difference(likely_regressions)
                & groups_high.union(groups_classified_as_new)
            )
            - groups_non_consistent_failure
        )

        # Sorting groups to be backfilled, first ones will be groups present in groups_consistent_failure with a high confidence
        failures_to_be_backfilled.sort(
            key=lambda group: (
                int(group in groups_consistent_failure)
                + bugbug_selection["groups"].get(group, 0)
            )
        )
        failures_to_be_backfilled.reverse()

        # Output real, intermittent and unknown groupfailures
        # along with their failing configurations + groupfailures to retrigger/backfill
        return Regressions(
            real=_map_failing_tasks(real_failures),
            intermittent=_map_failing_tasks(intermittent_failures),
            unknown=_map_failing_tasks(unknown_failures),
        ), ToRetriggerOrBackfill(
            real_retrigger=_map_failing_tasks(real_failures_to_be_retriggered),
            intermittent_retrigger=_map_failing_tasks(
                intermittent_failures_to_be_retriggered
            ),
            backfill=_map_failing_tasks(failures_to_be_backfilled),
        )

    def has_intermittent_bug(self, term: str, bugs: list[dict]) -> bool:
        """
        Identify whether an intermittent bug has be opened already based on a failure term, structured as below:
        TEST-UNEXPECTED-FAIL | path/name.ext | subtest name and/or further information
        """
        values = term.split("|")
        if len(values) != 3:
            logger.warning(
                f"Invalid failure term '{term}': considering the failure as new."
            )
            return False
        _, full_path, info = values
        # Consider that an intermittent bug exist for this issue based on its path + status
        return any(
            bug["status"] in ("NEW", "REOPENED")
            and bug["summary"].endswith(f"{full_path} | single tracking bug")
            for bug in bugs
        )

    def identify_permanent_failures(self, retriggers_count=5, queue_prefix="") -> None:
        """
        Retrigger failed test tasks `retriggers_count` times to identify the permanent failures.
        Only tasks with a Task Queue ID starting by the given prefix will be retriggered.
        Failures that are not considered permanent are backfilled to identify the push which caused the regression.
        """
        failures_to_retrigger = [
            task
            for task in self.tasks
            if (
                task.queue_id
                and task.queue_id.startswith(queue_prefix)
                and task.can_retrigger_test_failure()
            )
        ]
        for failure in failures_to_retrigger:
            # Detect the actual number of retriggers to plan new ones
            retriggers = [
                task
                for task in self.tasks
                if task.is_retrigger and task.label == failure.label
            ]
            # Look for existing retrigger action among tasks for this failure
            retrigger_actions = [
                retrigger
                for retrigger in self.tasks
                if (
                    retrigger.action
                    and retrigger.action.get("name") == "retrigger-multiple"
                    and failure.label
                    in itertools.chain(
                        request.get("tasks")
                        for request in retrigger.action.get("requests", [])
                    )
                    # Only count actions that did not really triggered a task yet
                    and not any(task.parent == retrigger.id for task in self.tasks)
                )
            ]
            # Count actions for which no task has been created yet
            scheduled_retriggers = sum(
                t.action["requests"].get("times", 0) for t in retrigger_actions
            )
            if scheduled_retriggers:
                logger.info(
                    f"{len(scheduled_retriggers)} retrigger actions have been detected "
                    f"for failure {failure.label} ({failure.id})"
                )

            to_retrigger = retriggers_count - len(retriggers) - scheduled_retriggers
            if to_retrigger <= 0:
                if not scheduled_retriggers:
                    # All tasks have been retriggered, check the results
                    self.backfill_new_test_failure(failure, retriggers)
                continue
            logger.info(
                f"Starting {to_retrigger} retriggers for task {failure.label} ({failure.id})"
            )
            failure.retrigger(self, times=to_retrigger)

    def backfill_new_test_failure(self, failure: Task, retriggers: list[Task]) -> None:
        """Analyze the result of the test retriggers to classify between backfills and permanent failures."""
        if any(
            retrigger.state in ("unscheduled", "pending", "running")
            for retrigger in retriggers
        ):
            logger.warning(
                f"Some retrigger did not finished for task {failure.label} ({failure.id}), skipping."
            )
            return

        # Aggregate all failure lines from the initial task and its retriggers, based on Treeherder logs parsing
        treeherder_client = TreeherderClientSource()
        overall_failure_lines: list[str] = []
        for task in [failure, *retriggers]:
            try:
                job = treeherder_client.get_job_from_task(task)
                treeherder_client.check_job_ready(job["id"], branch=self.branch)
            except JobUnavailable:
                # Ensure all the logs have been parsed in Treeherder before continuing
                return

            # Rely on the bug_suggestions data from Treeherder to determine the number of similar failures among retriggers
            bug_suggestions = treeherder_client.get_bug_suggestions(job["id"])
            if not bug_suggestions:
                continue
            terms = {
                suggestions["search"]: suggestions["bugs"]["open_recent"]
                for suggestions in bug_suggestions
                if suggestions["search"].startswith("TEST-UNEXPECTED-")
            }
            if not terms and task == failure:
                # The initial failure should return at least 1 valid error line to continue with the analysis
                logger.warning(
                    "No failure line in standard format with TEST-UNEXPECTED- found for initial task "
                    f"{failure.label} ({failure.id}) in the corresponding Treeherder job {job['id']}, skipping."
                )
                return
            elif not terms:
                logger.debug(
                    f"No failure line in standard format with TEST-UNEXPECTED- found for job {job['id']}"
                )
                continue

            # Filter out failures that already have an intermittent bug opened
            # to avoid backfills of known frequent intermittent failures.
            terms = {
                term: open_recent
                for term, open_recent in terms.items()
                if not self.has_intermittent_bug(term, open_recent)
            }
            if not terms:
                logger.debug("All the failures already have a bug opened, skipping.")
                continue
            overall_failure_lines.extend(set(terms))

        # Analyze the most common failure lines across the initial task and the retriggers
        most_common_failure = Counter(overall_failure_lines).most_common(1)

        value, counter = most_common_failure[0]
        logger.info(
            f"The most common failure line '{value}' has been observed {counter} "
            "times among initial task and its retriggers"
        )
        if counter >= 6:
            # All of the tasks failed with similar failure lines, backfill with one task per push
            logger.info(
                f"Test failure {failure.label} ({failure.id}): "
                "Backfilling the initial task as permanent, because all of retriggers seem "
                "to have caused the same failure."
            )
            failure.backfill(self, times=1)
        elif counter >= 3:
            # >=50% tasks failed (initial + 2 out of 5 retriggers), backfill with 6 tasks per push
            logger.info(
                f"Test failure {failure.label} ({failure.id}): "
                "Backfilling the initial task as intermittent, because >= 50% of the tasks seem "
                f"to have caused the same failure ({counter} out of {len(retriggers) + 1})."
            )
            failure.backfill(self, times=6)
        elif counter >= 2:
            logger.warning(
                f"Failure {failure.label} ({failure.id}): "
                f"found {counter} similar failure lines that seem to have caused the same failure."
            )
        else:
            logger.info(
                f"Failure {failure.label} ({failure.id}): "
                "No similar failure lines were found among the retriggers, there is nothing to do."
            )

    def classify(
        self,
        intermittent_confidence_threshold: float = 0.8,
        real_confidence_threshold: float = 0.9,
        use_possible_regressions: bool = False,
        unknown_from_regressions: bool = True,
        consider_children_pushes_configs: bool = True,
        cross_config_counts: Optional[Tuple[int, int]] = (2, 2),
        consistent_failures_counts: Optional[Tuple[int, int]] = (2, 3),
    ) -> Tuple[PushStatus, Regressions, ToRetriggerOrBackfill]:
        """
        Classify the overall push state using its group tasks states
        from classify_regressions:
        - bad push: when there are any task group with real failures
        - good push: when there are only intermittent failures
        - unknown state: when other tasks are failing
        """
        regressions, to_retrigger_or_backfill = self.classify_regressions(
            intermittent_confidence_threshold,
            real_confidence_threshold,
            use_possible_regressions,
            unknown_from_regressions,
            consider_children_pushes_configs,
            cross_config_counts,
            consistent_failures_counts,
        )

        # If there are any real failures, it's a bad push
        if len(regressions.real) > 0:
            return PushStatus.BAD, regressions, to_retrigger_or_backfill

        # If all failures are intermittent, it's a good push
        if len(regressions.unknown) == 0 and len(regressions.intermittent) >= 0:
            return PushStatus.GOOD, regressions, to_retrigger_or_backfill

        # Fallback to unknown
        return PushStatus.UNKNOWN, regressions, to_retrigger_or_backfill

    @memoize
    def get_shadow_scheduler_tasks(self, name: str) -> List[dict]:
        """Returns all tasks the given shadow scheduler would have scheduled,
        or None if the given scheduler didn't run.

        Args:
            name (str): The name of the shadow scheduler to query.

        Returns:
            set: All task labels that would have been scheduled.
        """
        index = self.index + ".source.shadow-scheduler-{}".format(name)
        task = Task.create(index=index)

        optimized = task.get_artifact("public/shadow-scheduler/optimized-tasks.json")
        return list(optimized.values())

    @property
    def shadow_scheduler_names(self) -> List[str]:
        return sorted(
            label.split("shadow-scheduler-")[1]
            for label in self.scheduled_task_labels
            if "shadow-scheduler" in label
        )

    def generate_all_shadow_scheduler_tasks(
        self,
    ) -> Iterator[Tuple[str, Union[Set[str], Exception]]]:
        """Generates all tasks from all of the shadow schedulers that ran on the push.

        Yields:
            tuple: Of the form (<name>, {<label>}) where the first value is the
            name of the shadow scheduler and the second is the set of tasks it
            would have scheduled, or an exception instance in case the shadow
            scheduler failed.
        """
        for name in sorted(self.shadow_scheduler_names):
            try:
                yield (
                    name,
                    set(t["label"] for t in self.get_shadow_scheduler_tasks(name)),
                )
            except Exception as e:
                yield name, e

    def generate_all_shadow_scheduler_config_groups(
        self,
    ) -> Iterator[Tuple[str, Union[Set[Tuple[str, str]], Exception]]]:
        """Generates all groups from all tasks from all of the shadow schedulers that
           ran on the push.

        Yields:
            tuple: Of the form (<name>, {(<config>, <group>)}) where the first value
            is the name of the shadow scheduler and the second is the set of groups, on
            given configurations, it would have scheduled, or an exception instance in
            case the shadow scheduler failed.
        """
        for name in sorted(self.shadow_scheduler_names):
            try:
                config_groups = {
                    (
                        get_configuration(task),
                        group,
                    )
                    for task in self.get_shadow_scheduler_tasks(name)
                    for group in task["attributes"].get("test_manifests", [])
                }

                yield name, config_groups
            except Exception as e:
                yield name, e

    @memoize
    def get_test_selection_data(self):
        """Retrieves Push schedules from various data sources.

        As for now, the 'push_test_selection_data' contract is fulfilled by two data sources:
        bugbug -> The call will be retried until we receive data or it times out/fails.
        taskcluster -> The call will be made one time.
        """
        return data.handler.get(
            "push_test_selection_data", branch=self.branch, rev=self.rev
        )

    def get_existing_classification(self, environment: str) -> PushStatus:
        """Retrieves existing classification from Taskcluster artifacts

        Do not memoize this method as the classification may change every few minutes remotely
        """
        existing = data.handler.get(
            "push_existing_classification",
            branch=self.branch,
            rev=self.rev,
            environment=environment,
        )

        # Convert from raw string to enum
        return PushStatus[str(existing)]

    def __repr__(self):
        return f"{super(Push, self).__repr__()} rev='{self.rev}'"


def make_push_objects(**kwargs):
    try:
        if "from_date" in kwargs and "to_date" in kwargs:
            # Load by date range
            pushes_data = data.handler.get("push_revisions", **kwargs)
        elif "nb" in kwargs:
            # Load latest pushes
            pushes_data = data.handler.get("pushes", **kwargs)
        else:
            raise Exception(
                "Unsupported parameters (either from_date and to_date or nb are required)"
            )
    except MissingDataError:
        return []

    pushes = []

    for push_data in pushes_data:
        extra = {}
        if "branch" in kwargs:
            extra = {"branch": kwargs["branch"]}

        cur = Push(push_data["revs"], **extra)

        # avoids the need to query hgmo to find this info
        cur._id = push_data["pushid"]
        cur._date = push_data["date"]

        pushes.append(cur)

    pushes.sort(key=lambda p: p._id)

    for i, cur in enumerate(pushes):
        if i != 0:
            cur._parent = pushes[i - 1]

        if i != len(pushes) - 1:
            cur._child = pushes[i + 1]

    return pushes


def make_summary_objects(from_date, to_date, branch, type):
    """Returns a list of summary objects matching the parameters.

    When invoked with a `type` argument, this method will return a list of
    Summary objects of the corresponding type. For example, `type='group'` will
    return a list of GroupSummary objects.

    Args:
        from_date (str): String representing a date value.
        to_date (str): String representing a date value.
        branch (str): String that references one of the Mozilla CI's repositories.
        type (str): String that references one of the supported Summary types.

    Returns:
        list: List of Summary objects, or an empty list.

    """
    # Retrieve the function by name using the provided `type` argument.
    func = "__make_{}_summary_objects".format(type).lower()

    # If the method name `func` (maps to either of the private methods)
    # is not found, then return an empty list.
    if not func:
        return []

    # Obtain list of all pushes for the specified branch, from_date and to_date.
    pushes = make_push_objects(from_date=from_date, to_date=to_date, branch=branch)

    # Flatten list of tasks for every push to a 1-dimensional list.
    tasks = sorted(
        [task for push in pushes for task in push.tasks], key=lambda x: x.label
    )

    summaries = globals()[func](tasks)

    # Sort by either the label (LabelSummary) or name (GroupSummary).
    summaries.sort(key=lambda x: getattr(x, "label", "name"))
    return summaries


def __make_label_summary_objects(tasks):
    """Generates a list of LabelSummary objects from a list of tasks.

    Args:
        tasks (list): List of Task objects.

    Returns:
        list: List of LabelSummary objects.
    """
    # Make a mapping keyed by task.label containing list of Task objects.
    tasks_by_config = defaultdict(lambda: defaultdict(list))
    for task in tasks:
        tasks_by_config[task.configuration][task.label]["tasks"].append(task)

    return [
        LabelSummary(label, tasks)
        for config in tasks_by_config.keys()
        for label, tasks in tasks_by_config[config].items()
    ]


def __make_group_summary_objects(tasks):
    """Generates a list of GroupSummary objects from a list of tasks.

    Args:
        tasks (list): List of Task objects.

    Returns:
        list: List of GroupSummary objects.
    """
    # Make a mapping keyed by group name containing list of Task objects.
    tasks_by_config = defaultdict(lambda: defaultdict(list))
    for task in tasks:
        for result in task.results:
            # Build the mapping of group to the group durations and TestTask objects.
            tasks_by_config[task.configuration][result.group].append(task)

    return [
        GroupSummary(key, tasks)
        for config in tasks_by_config.keys()
        for key, tasks in tasks_by_config[config].items()
    ]



# -*- coding: utf-8 -*-

import sys

from cleo.application import Application

from mozci.console.commands.batch_execution import (
    BatchClassificationCommand,
    BatchEvaluationCommand,
)
from mozci.console.commands.build_regression import RegressionCommand
from mozci.console.commands.check_backfills import CheckBackfillsCommand
from mozci.console.commands.decision import DecisionCommand
from mozci.console.commands.push import (
    ClassifyCommand,
    ClassifyEvalCommand,
    ClassifyPerfCommand,
    PushTasksCommand,
)


def cli():
    application = Application()
    application.add(BatchClassificationCommand())
    application.add(BatchEvaluationCommand())
    application.add(CheckBackfillsCommand())
    application.add(ClassifyCommand())
    application.add(ClassifyEvalCommand())
    application.add(ClassifyPerfCommand())
    application.add(PushTasksCommand())
    application.add(DecisionCommand())
    application.add(RegressionCommand())
    application.run()


if __name__ == "__main__":
    sys.exit(cli())



# -*- coding: utf-8 -*-
import csv
import datetime
import itertools
import json
import os
import time
import uuid
from multiprocessing import Pool

from cleo.commands.command import Command
from cleo.helpers import option
from loguru import logger

from mozci.console.commands.push import (
    parse_and_log_details,
    prepare_for_analysis,
    retrieve_sheriff_intermittents,
    retrieve_sheriff_reals,
)
from mozci.push import Push, PushStatus, make_push_objects
from mozci.util.defs import INTERMITTENT_CLASSES

BASE_OUTPUT_DIR = (
    os.path.dirname(os.path.realpath(__file__)) + "/classify_batch_results"
)
DAYS_FROM_TODAY = 30
PARAMETERS_NAMES = [
    "intermittent_confidence_threshold",
    "real_confidence_threshold",
    "use_possible_regressions",
    "unknown_from_regressions",
    "consider_children_pushes_configs",
    "cross_config_counts",
    "consistent_failures_counts",
]
PARAMETERS_COMBINATIONS = [
    dict(zip(PARAMETERS_NAMES, parameters_values))
    for parameters_values in itertools.product(
        [0.5, 0.7, 0.9],  # intermittent_confidence_threshold
        [0.7, 0.8, 0.9],  # real_confidence_threshold
        [True, False],  # use_possible_regressions
        [True, False],  # unknown_from_regressions
        [True, False],  # consider_children_pushes_configs
        [(2, 2), (2, 3), (2, 5)],  # cross_config_counts
        [(2, 3), (2, 5)],  # consistent_failures_counts
    )
]
MAXIMUM_PROCESSES = os.cpu_count() or 1


def retrieve_push_and_prepare_for_analysis(push):
    push = Push(push["rev"], branch=push["branch"])
    prepare_for_analysis(push)


def _serialize_regressions(regressions):
    return {
        group: [{"task_id": task.id, "label": task.label} for task in failing_tasks]
        for group, failing_tasks in regressions.items()
    }


def create_json_file(push, run_id, classification_name, regressions):
    to_save = {
        "push": {
            "id": push.push_uuid,
            "classification": classification_name,
        },
        "failures": {
            "real": _serialize_regressions(regressions.real),
            "intermittent": _serialize_regressions(regressions.intermittent),
            "unknown": _serialize_regressions(regressions.unknown),
        },
    }

    with open(f"{BASE_OUTPUT_DIR}/{push.id}/{run_id}.json", "w") as file:
        json.dump(to_save, file, indent=2)


def run_combinations_for_push(push):
    push = Push(push["rev"], branch=push["branch"])

    push_dir = f"{BASE_OUTPUT_DIR}/{push.id}"
    if not os.path.exists(push_dir):
        os.makedirs(push_dir)

    csv_rows = []
    for parameters in PARAMETERS_COMBINATIONS:
        run_id = uuid.uuid4()

        start = time.time()
        try:
            classification, regressions = push.classify(**parameters)
            # Rettriger test failures to detect permanent ones, only on Linux cloud worker pool for now
            push.identify_permanent_failures(queue_prefix="gecko-t/t-linux-")
            end = time.time()

            # Only save results to a JSON file if the execution was successful
            classification_name = classification.name
            create_json_file(push, run_id, classification_name, regressions)
        except Exception as e:
            end = time.time()
            classification_name = "SYSTEM_ERROR"
            logger.error(
                f"An error occurred during the classification of push {push.push_uuid}: {e}"
            )

        csv_rows.append(
            {
                "run_uuid": run_id,
                "push_uuid": push.push_uuid,
                **parameters,
                "classification": classification_name,
                "time_spent": round(end - start, 3),
                "now": datetime.datetime.now(),
            }
        )

    return csv_rows


class BatchClassificationCommand(Command):
    name = "batch-execution classify"
    description = "Run the classification algorithm with various parameters combinations for all submitted pushes within the last 30 days"
    options = [
        option(
            "workers",
            description="Number of workers to use in order to parallelize the executions.",
            flag=False,
        )
    ]

    csv_header = (
        ["run_uuid", "push_uuid"]
        + PARAMETERS_NAMES
        + ["classification", "time_spent", "now"]
    )

    def handle(self) -> None:
        # Default value from a constant if nothing is provided
        workers_count = MAXIMUM_PROCESSES

        workers_option = self.option("workers")
        if workers_option:
            try:
                workers_count = int(workers_option)
            except ValueError:
                self.line("<error>Provided --workers should be an int.</error>")
                exit(1)

            if workers_count > MAXIMUM_PROCESSES:
                self.line(
                    f"<comment>Parallelization over {workers_count} workers was requested but only {MAXIMUM_PROCESSES} CPUs are available, falling back to using only {MAXIMUM_PROCESSES} workers.</comment>"
                )
                workers_count = MAXIMUM_PROCESSES

        pushes = self.retrieve_pushes()
        self.line(
            f"<info>{len(pushes)} pushes will be classified using {len(PARAMETERS_COMBINATIONS)} parameters combinations.</info>"
        )

        with Pool(workers_count) as pool:
            # Populate the cache + Clean up potential biases
            pool.map(retrieve_push_and_prepare_for_analysis, pushes)

        if not os.path.exists(BASE_OUTPUT_DIR):
            os.makedirs(BASE_OUTPUT_DIR)

        with open(
            BASE_OUTPUT_DIR + "/all_executions.csv", "w", encoding="UTF8", newline=""
        ) as f:
            writer = csv.DictWriter(f, fieldnames=self.csv_header)
            writer.writeheader()

            with Pool(workers_count) as pool:
                # Each time an execution ends (execution = all combinations for a single push),
                # its result will be appended to the CSV
                for csv_rows in pool.imap(run_combinations_for_push, pushes):
                    writer.writerows(csv_rows)

    def retrieve_pushes(self):
        now = datetime.datetime.today()
        to_date = now.strftime("%Y-%m-%d")
        from_date = (now - datetime.timedelta(days=DAYS_FROM_TODAY)).strftime(
            "%Y-%m-%d"
        )
        pushes = make_push_objects(
            from_date=from_date, to_date=to_date, branch="autoland"
        )
        return [{"rev": push.rev, "branch": push.branch} for push in pushes]


class BatchEvaluationCommand(Command):
    name = "batch-execution evaluate"
    description = "Evaluate and aggregate the results produced by the classify_batch_execution script"

    csv_header = [
        "configuration",
        "total_pushes",
        "classify_errors",
        "evaluate_errors",
        "correct_good",
        "wrong_good",
        "correct_bad",
        "wrong_bad",
        "total_real",
        "correct_real",
        "wrong_real",
        "missed_real",
        "total_intermittent",
        "correct_intermittent",
        "wrong_intermittent",
        "missed_intermittent",
    ]

    def handle(self) -> None:
        results_path = BASE_OUTPUT_DIR + "/all_executions.csv"
        assert os.path.exists(results_path) and os.path.isfile(results_path), (
            "The CSV containing results from all classify executions doesn't exist"
        )

        with open(results_path, newline="") as f:
            reader = csv.DictReader(f)
            configs = self.parse_csv(reader)

        # Write evaluation results per config in a new CSV file
        with open(
            BASE_OUTPUT_DIR + "/evaluation_of_all_executions.csv",
            "w",
            encoding="UTF8",
            newline="",
        ) as f:
            writer = csv.DictWriter(f, fieldnames=self.csv_header)
            writer.writeheader()

            csv_rows = []
            for config, results in configs.items():
                results.update({"configuration": config})
                csv_rows.append(results)

            writer.writerows(csv_rows)

    def evaluate_push_failures(
        self, push_id, run_uuid, push_group_summaries, sheriff_data
    ):
        """
        Evaluate real/intermittent failures on a Push
        """
        with open(f"{BASE_OUTPUT_DIR}/{push_id}/{run_uuid}.json") as f:
            classify_results = json.load(f)
            predicted_reals = classify_results["failures"]["real"].keys()
            predicted_intermittents = classify_results["failures"][
                "intermittent"
            ].keys()

        evaluation = {}
        evaluation.update(
            # Evaluate real failures that were predicted
            parse_and_log_details(
                push_group_summaries,
                sheriff_data["reals"],
                {"fixed by commit"},
                predicted_groups=predicted_reals,
                ignore_pending_conflicting=True,
                suffix="_real",
            )[0]
        )
        evaluation.update(
            # Evaluate intermittent failures that were predicted
            parse_and_log_details(
                push_group_summaries,
                sheriff_data["intermittents"],
                set(INTERMITTENT_CLASSES),
                predicted_groups=predicted_intermittents,
                ignore_pending_conflicting=True,
                suffix="_intermittent",
            )[0]
        )

        return evaluation

    def parse_csv(self, reader):
        """
        Parse the CSV file and evaluate each classify result for a specific configuration and Push
        """
        pushes_group_summaries = {}
        pushes_sheriff_data = {}

        configs = {}
        for i, row in enumerate(reader, start=2):
            self.line(f"Evaluating line {i} of the CSV")
            config = ",".join(
                ["=".join([param, row[param]]) for param in PARAMETERS_NAMES]
            )

            if config not in configs:
                # Initialize all counts at 0
                configs[config] = {key: 0 for key in self.csv_header[1:]}

            configs[config]["total_pushes"] += 1

            if row["classification"] == "SYSTEM_ERROR":
                # No need to evaluate since an error happened during Push classification
                configs[config]["classify_errors"] += 1
                continue

            push_uuid = row["push_uuid"]
            try:
                branch, rev = push_uuid.split("/")
                push = Push(rev, branch=branch)

                if row["classification"] == PushStatus.GOOD.name:
                    if push.backedout:
                        configs[config]["wrong_good"] += 1
                    else:
                        configs[config]["correct_good"] += 1
                elif row["classification"] == PushStatus.BAD.name:
                    if push.backedout:
                        configs[config]["correct_bad"] += 1
                    else:
                        configs[config]["wrong_bad"] += 1

                if push_uuid not in pushes_group_summaries:
                    pushes_group_summaries[push_uuid] = push.group_summaries

                if push_uuid not in pushes_sheriff_data:
                    # By storing sheriffs annotations here, we will only have to run
                    # retrieve_sheriff_data once per push, since it can take a long
                    # time to fetch everything, this helps speed up the evaluation
                    pushes_group_summaries, sheriff_reals = retrieve_sheriff_reals(
                        pushes_group_summaries, push
                    )
                    (
                        pushes_group_summaries,
                        sheriff_intermittents,
                    ) = retrieve_sheriff_intermittents(pushes_group_summaries, push)
                    pushes_sheriff_data[push_uuid] = {
                        "reals": sheriff_reals,
                        "intermittents": sheriff_intermittents,
                    }

                evaluation = self.evaluate_push_failures(
                    push.id,
                    row["run_uuid"],
                    pushes_group_summaries[push_uuid],
                    pushes_sheriff_data[push_uuid],
                )
                for key, value in evaluation.items():
                    configs[config][key] += value

            except Exception as e:
                configs[config]["evaluate_errors"] += 1
                self.line(
                    f"<error>An error occurred during the evaluation of the classify execution with run_uuid {row['run_uuid']}: {e}</error>"
                )

        return configs



# -*- coding: utf-8 -*-

FAILURE_CLASSES = ("not classified", "new failure not classified", "fixed by commit")
INTERMITTENT_CLASSES = (
    "intermittent",
    "autoclassified intermittent",
    "intermittent needs bugid",
)
TASK_FINAL_STATES = ("completed", "failed", "exception")
